# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import datetime as dt
from collections import OrderedDict
import jwt

from sqlalchemy import (Column, Integer, String, JSON,
                        ForeignKey, Text, Table, Float)
from sqlalchemy.orm import relationship
from sqlalchemy_utc import UtcDateTime, utcnow
from flask_login import UserMixin
from flask import current_app as app
from werkzeug.security import generate_password_hash, check_password_hash


def create_data_models(dm):
    """ Define the Data Models that will be use by the DataManager. """

    Base = dm.Base


    class Resource(Base):
        """ Representation of different type of Resources.
        (e.g Microscopes, other instruments or services.
        """
        __tablename__ = 'resources'

        id = Column(Integer,
                    primary_key=True)
        name = Column(String(64),
                      index=True,
                      unique=True,
                      nullable=False)

        # Possible statuses of a Resource:
        #   - pending: when then entry is created but not visible
        #   - active: it has been activated for operation
        #   - inactive: it has been closed and it becomes inactive
        status = Column(String(32), default='active')

        tags = Column(String(256),
                      nullable=False)

        image = Column(String(64),
                       nullable=False)

        color = Column(String(16),
                       nullable=False)

        # General JSON dict to store extra attributes
        extra = Column(JSON, default={})

        def json(self):
            return dm.json_from_object(self)

        def __getExtra(self, key, default):
            return self.extra.get(key, default)

        def __setExtra(self, key, value):
            extra = dict(self.extra or {})
            extra[key] = value
            self.extra = extra

        @property
        def requires_slot(self):
            """ If True, users will need to have access to an slot
            or have some exceptions via the Application.
            """
            return  self.__getExtra('requires_slot', False)

        @requires_slot.setter
        def requires_slot(self, value):
            self.__setExtra('requires_slot', value)

        # Latest number of hours that a booking can be canceled
        # for this resource (e.g until 48h for a booking on Krios)
        # If 0, means that the booking can be cancelled at any time
        @property
        def latest_cancellation(self):
            return  self.__getExtra('latest_cancellation', 0)

        @latest_cancellation.setter
        def latest_cancellation(self, value):
            self.__setExtra('latest_cancellation', int(value))

        @property
        def min_booking(self):
            """ Minimum amount of hours that should be used for
            booking this resource.
            """
            return self.__getExtra('min_booking', 0)

        @min_booking.setter
        def min_booking(self, value):
            self.__setExtra('min_booking', int(value))

        @property
        def max_booking(self):
            """ Maximum amount of hours that can be used in a booking for
            booking this resource.
            """
            return self.__getExtra('max_booking', 0)

        @max_booking.setter
        def max_booking(self, value):
            self.__setExtra('max_booking', int(value))

        @property
        def is_microscope(self):
            return 'microscope' in self.tags

        @property
        def is_active(self):
            """ Return True if this instrument is "active".
            If not "active" bookings can not be made for this instrument.
            """
            return self.status == 'active'

        @property
        def requires_application(self):
            """ True if the user should belongs to an Application to book this resource.
            """
            return self.__getExtra('requires_application', True)

        @requires_application.setter
        def requires_application(self, value):
            self.__setExtra('requires_application', bool(value))

        @property
        def daily_cost(self):
            """ Cost of one day session using this resource.
            """
            return self.__getExtra('daily_cost', 0)

        @daily_cost.setter
        def daily_cost(self, value):
            self.__setExtra('daily_cost', int(value))


    ApplicationUser = Table('application_user', Base.metadata,
                            Column('application_id', Integer,
                                   ForeignKey('applications.id')),
                            Column('user_id', Integer,
                                   ForeignKey('users.id')))


    class User(UserMixin, Base):
        """Model for user accounts."""
        __tablename__ = 'users'
        ROLES = ['user', 'admin', 'manager', 'head', 'pi', 'independent',
                 'staff-solna', 'staff-umea']

        STATUSES = ['pending', 'active', 'inactive']

        id = Column(Integer,
                    primary_key=True)
        username = Column(String(64),
                          index=True,
                          unique=True,
                          nullable=False)
        email = Column(String(80),
                       index=False,
                       unique=True,
                       nullable=False)

        phone = Column(String(80))

        name = Column(String(256),
                      nullable=False)

        created = Column(UtcDateTime,
                         index=False,
                         unique=False,
                         nullable=False,
                         default=utcnow())

        # Possible statuses of a User:
        #   - pending: when then account is created and pending approval
        #   - active: user is ready for operation
        #   - inactive: user is not not longer active
        status = Column(String(32), default='active')

        # Default role should be: 'user'
        # more roles can defined in a json list: ['user', 'admin', 'manager', 'pi']
        roles = Column(JSON, nullable=False, default=['user'])

        password_hash = Column(String(256),
                               unique=True,
                               nullable=False)

        profile_image = Column(String(256), unique=True, nullable=True)

        # ---------------- RELATIONS ----------------------------

        # Pi and Lab_members relation
        lab_members = relationship("User", back_populates='pi')

        pi_id = Column(Integer, ForeignKey('users.id'),
                       nullable=True)
        pi = relationship("User",
                          foreign_keys=[pi_id],
                          remote_side=[id],
                          back_populates="lab_members")

        # one user to many sessions, bidirectional
        sessions = relationship('Session', back_populates="operator")

        created_applications = relationship("Application",
                                            back_populates="creator")

        # many to many relation with Application
        applications = relationship("Application",
                                    secondary=ApplicationUser,
                                    back_populates="users")

        transactions = relationship('Transaction', back_populates="user")

        #projects = relationship('Project', back_populates="user")

        # General JSON dict to store extra attributes
        extra = Column(JSON, default={})

        @staticmethod
        def create_password_hash(password):
            return generate_password_hash(password, method='sha256')

        def set_password(self, password):
            """Create hashed password."""
            self.password_hash = self.create_password_hash(password)

        def check_password(self, password):
            """Check hashed password."""
            return check_password_hash(self.password_hash, password)

        def get_reset_password_token(self, expires_in=600):
            token = jwt.encode(
                {'reset_password': self.id,
                 'exp': dm.now() + dt.timedelta(seconds=expires_in)},
                app.config['SECRET_KEY'], algorithm='HS256').decode('utf-8')
            self.extra = {'reset_token': token}

            return token

        @staticmethod
        def verify_reset_password_token(token):
            try:
                user_id = jwt.decode(token, app.config['SECRET_KEY'],
                                algorithms=['HS256'])['reset_password']
            except:
                return None
            user = User.query.get(user_id)

            if user is not None:
                if token != user.extra.get('reset_token', None):
                    return None
                # Valid token, let's make it invalid from now
                user.extra = {'reset_token': None}

            return user

        def __repr__(self):
            return '<User {}>'.format(self.username)

        def json(self):
            return dm.json_from_object(self)

        @property
        def is_developer(self):
            return 'developer' in self.roles

        @property
        def is_admin(self):
            return 'admin' in self.roles or self.is_developer

        @property
        def is_manager(self):
            return 'manager' in self.roles or self.is_admin or self.is_head

        @property
        def is_head(self):
            return 'head' in self.roles

        @property
        def is_staff(self):
            return any(r.startswith('staff-') for r in self.roles)

        @property
        def staff_unit(self):
            for r in self.roles:
                if r.startswith('staff-'):
                    return r.replace('staff-', '')
            return None

        @property
        def is_pi(self):
            return 'pi' in self.roles

        @property
        def is_independent(self):
            return 'independent' in self.roles

        @property
        def is_application_manager(self):
            return len(self.created_applications) > 0

        @property
        def is_active(self):
            return self.status == 'active'

        @property
        def rolesmap(self):
            return {role: role in self.roles
                    for role in self.ROLES}

        def get_pi(self):
            """ Return the PI of this user. PI are consider PI of themselves.
            """
            return self if self.is_pi else self.pi

        def same_pi(self, other):
            """ Return if the same pi. """
            return other is not None and self.get_pi() == other.get_pi()

        def get_applications(self, status='active'):
            """ Return the applications of this user.
            """
            applications = []
            pi = self.get_pi()
            if pi is not None:
                applications = list(pi.created_applications)
                for a in pi.applications:
                    if a not in applications:
                        applications.append(a)
                #applications.extend(pi.applications)

            def _filter(a):
                return status == 'all' or a.status == status

            return [a for a in applications if _filter(a)]

        def has_application(self, applicationCode):
            """ Return True if the user has the given application. """
            return any(a.code == applicationCode
                       for a in self.get_applications(status='all'))

        def get_lab_members(self, onlyActive=True):
            """ Return lab members, filtering or not by active status. """
            if onlyActive:
                return [u for u in self.lab_members if u.is_active]
            else:
                return self.lab_members

        def can_book_resource(self, resource):
            """ Return  True if the user can book a given resource without
            an  explicit SLOT for it. """
            if self.is_manager or not resource.requires_slot:
                return True

            apps = self.get_applications()

            # If the user is not manager and the resource requires slot,
            # let's check if there is any application that allows the user
            # to book without a given SLOT
            return any(a.no_slot(resource.id) for a in apps)

        def can_book_slot(self, booking_slot):
            """ Return True if the user can book in the given SLOT. """
            return booking_slot.allows_user_in_slot(self)


    class Template(Base):
        """ Classes used as template to create Applications.
        Template instances that are 'active' will allow to
        create new applications from it. Otherwise, templates can be
        closed and will not be visible when creating a new application. """
        __tablename__ = 'templates'

        id = Column(Integer,
                    primary_key=True)

        # Possible statuses of a Template:
        #   - preparation: when it has been created and it under preparation
        #   - active: it has been activated for operation
        #   - closed: it has been closed and it becomes inactive
        status = Column(String(32), default='preparation')

        title = Column(String(256), nullable=False)

        description = Column(Text, nullable=True)

        # This will be data in json form to describe extra parameters defined
        # in this template for all the Applications created from this.
        form_schema = Column(JSON, nullable=True)

        applications = relationship("Application", back_populates='template')

        # General JSON dict to store extra attributes
        extra = Column(JSON, default={})

        def __getExtra(self, key, default):
            return self.extra.get(key, default)

        def __setExtra(self, key, value):
            extra = dict(self.extra)
            extra[key] = value
            self.extra = extra

        @property
        def code_prefix(self):
            """ Prefix for numbering Application Codes
            (e.g CEM, EXT, etc)
            """
            return  self.__getExtra('code_prefix', [])

        @code_prefix.setter
        def codes(self, value):
            self.__setExtra('code_prefix', value)

        def json(self):
            return dm.json_from_object(self)


    class Application(Base):
        """
        Application that applies for access to the facility.
        Usually many principal investigators are associated to a project.
        """
        __tablename__ = 'applications'

        id = Column(Integer,
                    primary_key=True)

        code = Column(String(32),
                      nullable=False,
                      unique=True)

        created = Column(UtcDateTime,
                         index=False,
                         unique=False,
                         nullable=False,
                         default=utcnow())

        alias = Column(String(32))

        # Possible statuses of an Application:
        #   - preparation: when it has been created and it under preparation
        #   - review: the application has been submitted for review
        #   - rejected: if for some reason the application is rejected
        #   - accepted: the application has been accepted after evaluation
        #   - active: it has been activated for operation
        #   - closed: it has been closed and it becomes inactive
        STATUSES = ['preparation', 'review', 'accepted', 'active', 'closed']
        status = Column(String(32), default='preparation')

        title = Column(String(256),
                       nullable=False)

        description = Column(Text,
                             nullable=True)

        invoice_reference = Column(String(256),
                                   nullable=False)

        invoice_address = Column(Text,
                                 nullable=True)

        DEFAULT_ALLOCATION = {'quota': {'krios': 0, 'talos': 0},
                              'noslot': []}
        # This is the maximum amount of days allocated per type of resource
        # {'krios': 20} means that this application can book max to 20 days
        # for any resource with tag 'krios' (i.e Titan Krios scopes)
        resource_allocation = Column(JSON, default=DEFAULT_ALLOCATION)

        # ID of the user that created the Application, it should be a PI
        creator_id = Column(Integer, ForeignKey('users.id'), nullable=False)
        creator = relationship("User", foreign_keys=[creator_id],
                               back_populates="created_applications")

        users = relationship("User",
                             secondary=ApplicationUser,
                             back_populates="applications")

        bookings = relationship("Booking",
                                back_populates="application")

        # Link to the template used to create the form
        template_id = Column(Integer, ForeignKey('templates.id'), nullable=False)
        template = relationship("Template", foreign_keys=[template_id],
                                back_populates="applications")

        # General JSON dict to store extra attributes
        extra = Column(JSON, default={})

        def __repr__(self):
            return '<Application code=%s, alias=%s>' % (self.code, self.alias)

        def __getExtra(self, key, default):
            return self.extra.get(key, default)

        def __setExtra(self, key, value):
            extra = dict(self.extra)
            extra[key] = value
            self.extra = extra

        @property
        def confidential(self):
            """ Return extra costs associated with this Booking
            """
            return self.__getExtra('confidential', False)

        @confidential.setter
        def costs(self, value):
            self.__setExtra('confidential', value)

        @property
        def access_list(self):
            """ Return a list of IDs with managers that
            can access this application. """
            return [ad['user_id'] for ad in self.__getExtra('access', [])]

        def allows_access(self, user):
            """ Return True if this user has access to the application. """
            if self.creator_id == user.id or user.is_admin:
                return True

            if user.is_manager:
                return not self.confidential or user.id in self.access_list

            user_apps = user.get_applications(status='all')
            return any(self.id == a.id for a in user_apps)

        def json(self):
            json = dm.json_from_object(self)
            json['pi_list'] = [pi.id for pi in self.pi_list]
            return json

        @property
        def is_active(self):
            return self.status == 'active'

        def get_quota(self, key):
            """ Return the quota (in days) for this application.
            Args:
                key: the key representing either a resource id or a tag (e.g 1 or 'krios')
            Return:
                the number of days that this application can use a resource or a groups
                resources with the same tag. It will return None if there is no quota for
                such resource.
            """
            return self.resource_allocation['quota'].get(key, None)

        def no_slot(self, resourceKey):
            """ Return True if this application is not forced to create bookings within
            a given slot for certain resource.
             For example, some applications might be required to book Krios1 only on specific
             slots, while others will be able to book in free days.
            """
            return resourceKey in self.resource_allocation['noslot']

        @property
        def pi_list(self):
            """ Return the list of PI. """
            # Since we are importing data now from the Portal, some application
            # have the creator of the application as one of the users, so we want
            # to avoid duplicated entries
            pi_list = []

            # Applications can also be created by facility staff, so only if the
            # creator is a pi we reported in the "pi_list"
            if self.creator is not None and self.creator.is_pi:
                pi_list.append(self.creator)

            for u in self.users:
                if u.id != self.creator.id:
                    pi_list.append(u)
            return pi_list


    class Booking(Base):
        """Model for user accounts."""
        __tablename__ = 'bookings'

        MAINTENANCE_LIST = ['cycle', 'installation', 'maintenance', 'afis']
        DEVELOPMENT_LIST = ['method', 'research', 'test', 'mikroed', 'microed', 'devel']

        id = Column(Integer, primary_key=True)

        title = Column(String(256), nullable=False)

        start = Column(UtcDateTime, nullable=False)

        end = Column(UtcDateTime, nullable=False)

        # booking, slot, downtime or maintenance
        type = Column(String(16), nullable=False)

        # slot authorization, who can book within this slot
        slot_auth = Column(JSON, default={'applications': [], 'users': []})

        description = Column(Text, nullable=True)

        repeat_id = Column(String(256), nullable=True)

        repeat_value = Column(String(32), nullable=False, default='no')

        resource_id = Column(Integer, ForeignKey('resources.id'))
        resource = relationship("Resource")

        # This is reference to the user that created the Booking
        creator_id = Column(Integer, ForeignKey('users.id'),
                            nullable=False)
        creator = relationship("User", foreign_keys=[creator_id])

        # And this is the user that "owns" the Booking
        owner_id = Column(Integer, ForeignKey('users.id'),
                          nullable=False)
        owner = relationship("User", foreign_keys=[owner_id])

        # And this is the user that "owns" the Booking
        operator_id = Column(Integer, ForeignKey('users.id'),
                          nullable=True)
        operator = relationship("User", foreign_keys=[operator_id])

        # Related to the Owner, we also keep the Application to which
        # this booking is associated
        application_id = Column(Integer, ForeignKey('applications.id'),
                                nullable=True)
        application = relationship("Application",
                                   back_populates="bookings")

        session = relationship("Session", back_populates="booking")

        # Experiment description
        experiment = Column(JSON, nullable=True)

        # General JSON dict to store extra attributes
        extra = Column(JSON, default={})

        @property
        def duration(self):
            return self.end - self.start

        @property
        def days(self):
            """ Count how many days these bookings spans.
            (It is not strictly necessary the total amount of time in in
            units of 24h.
            """
            td = self.end.date() - self.start.date() + dt.timedelta(days=1)
            return td.days

        @property
        def is_slot(self):
            return self.type == 'slot'

        def __repr__(self):
            def _timestr(dt):
                return dt.strftime('%Y/%m/%d')

            return ('<Booking: resource=%s, owner=%s, dates: %s - %s>'
                    % (self.resource.name, self.owner.name,
                       _timestr(self.start), _timestr(self.end)))

        def __getExtra(self, key, default):
            return self.extra.get(key, default)

        def __setExtra(self, key, value):
            extra = dict(self.extra)
            extra[key] = value
            self.extra = extra

        @property
        def costs(self):
            """ Return extra costs associated with this Booking
            """
            return  self.__getExtra('costs', [])

        @costs.setter
        def costs(self, value):
            self.__setExtra('costs', value)

        @property
        def total_cost(self):
            """ Return all costs associated with this Booking
            """
            cost = self.days * self.resource.daily_cost
            for _, _, c in self.costs:
                try:
                    cost += int(c)
                except:
                    pass
            return cost

        def json(self):
            return dm.json_from_object(self)

        def allows_user_in_slot(self, user):
            """ Return True if a given user is allowed to book in this Slot.
            """
            if not self.is_slot:
                return False

            if user.is_manager:
                return True

            allowedUsers = self.slot_auth.get('users', [])
            allowedApps = self.slot_auth.get('applications', [])

            return (user.id in allowedUsers or 'any' in allowedApps or
                    any(a.code in allowedApps for a in user.get_applications()))

        def application_in_slot(self, application):
            if not self.is_slot:
                return False

            return application.code in self.slot_auth.get('applications', [])


    class Session(Base):
        """Model for sessions."""
        __tablename__ = 'sessions'

        id = Column(Integer,
                    primary_key=True)

        name = Column(String(256),
                      unique=True,
                      nullable=False)

        start = Column(UtcDateTime)

        end = Column(UtcDateTime)

        # Possible statuses of a Session:
        #   - pending (stored in db, but data folders not created)
        #   - created (data folders created, but no processing reported)
        # TODO: review if the following states make sense, we might want to
        # TODO: decouple the session from the associated pre-processing
        #   - running
        #   - failed
        #   - finished
        status = Column(String(32), default='created')

        data_path = Column(String(256),
                           index=False,
                           nullable=True)

        DEFAULT_ACQUISITION = {
            'voltage': None,
            'cs': None,
            'phasePlate': False,
            'detector': None,
            'detectorMode': None,
            'pixelSize': None,
            'dosePerFrame': None,
            'totalDose': None,
            'exposureTime': None,
            'numOfFrames': None,
        }
        # Acquisition info parameters are store as a JSON string
        acquisition = Column(JSON, default=DEFAULT_ACQUISITION)

        DEFAULT_STATS = {
            'numOfMovies': 0,
            'numOfMics': 0,
            'numOfCtfs': 0,
            'numOfPtcls': 0,
            'numOfCls2D': 0,
            'ptclSizeMin': 0,
            'ptclSizeMax': 0,
        }

        # Acquisition info parameters are store as a JSON string
        stats = Column(JSON, default=DEFAULT_STATS)

        # Resource (usually microscope) that was used in this session
        # This should be the same resource as the booking, when it is not None
        resource_id = Column(Integer, ForeignKey('resources.id'))
        resource = relationship("Resource")

        # Booking that this Session is related (optional)
        booking_id = Column(Integer, ForeignKey('bookings.id'),
                            nullable=True)
        booking = relationship("Booking", back_populates="session")

        # User that was or is in charge of the session
        # It might be one of the facility staff or an independent user
        operator_id = Column(Integer, ForeignKey('users.id'), nullable=False)
        operator = relationship("User", back_populates="sessions")

        # General JSON dict to store extra attributes
        extra = Column(JSON, default={})

        class Cost:
            def __init__(self, id, date, comment, amount):
                self.id = id
                self.date = date
                self.comment = comment
                self.amount = amount

        @property
        def costs(self):
            return [self.Cost(i, dt.datetime.utcnow(), "Cost number %d" % i, i * 100)
                    for i in range(1, 4)]

        def __repr__(self):
            return '<Session {}>'.format(self.name)

        def json(self):
            return dm.json_from_object(self)


    class Form(Base):
        """ Class to store Forms definitions. """
        __tablename__ = 'forms'

        id = Column(Integer,
                    primary_key=True)

        name = Column(String(256),
                      unique=True,
                      nullable=False)

        # Form sections and params definition
        definition = Column(JSON, default={})

        def json(self):
            return dm.json_from_object(self)


    class InvoicePeriod(Base):
        """ Period for which invoices will be generated. """
        __tablename__ = 'invoice_periods'

        id = Column(Integer,
                    primary_key=True)

        start = Column(UtcDateTime,
                       nullable=False)

        end = Column(UtcDateTime,
                     nullable=False)

        # Possible statuses of an InvoicePeriod:
        #   - created
        #   - active
        #   - closed
        status = Column(String(32), default='active')

        # General JSON dict to store extra attributes
        extra = Column(JSON, default={})

        def __getExtra(self, key, default):
            return self.extra.get(key, default)

        def __setExtra(self, key, value):
            extra = dict(self.extra)
            extra[key] = value
            self.extra = extra

        def json(self):
            return dm.json_from_object(self)


    class Transaction(Base):
        """ Financial transactions. """
        __tablename__ = 'transactions'

        id = Column(Integer,
                    primary_key=True)

        date = Column(UtcDateTime,
                      nullable=False)

        amount  = Column(Float, default=0)

        comment = Column(String(256))

        # General JSON dict to store extra attributes
        extra = Column(JSON, default={})

        # User to which this Transaction is associated with
        # Usually it is a PI and the transaction is related to invoices
        user_id = Column(Integer, ForeignKey('users.id'), nullable=False)
        user = relationship("User", back_populates="transactions")

        def __getExtra(self, key, default):
            return self.extra.get(key, default)

        def __setExtra(self, key, value):
            extra = dict(self.extra)
            extra[key] = value
            self.extra = extra

        def json(self):
            return dm.json_from_object(self)


    class Project(Base):
        """ Project entity to group shipments, grids preparation,
         data collections and data processing.
         """
        __tablename__ = 'projects'

        id = Column(Integer, primary_key=True)

        date = Column(UtcDateTime, nullable=False)

        STATUS = ['active', 'inactive']
        # Possible statuses of a Project:
        #   - active: default state when created
        #   - inactive: it has been closed and it becomes inactive
        status = Column(String(32), default='active')

        title = Column(String(256), nullable=False)

        description = Column(Text, nullable=True)

        # User to which this Project is associated with
        # The PI of this project is inferred
        user_id = Column(Integer, ForeignKey('users.id'), nullable=False)
        user = relationship("User", #back_populates= "projects",
                            foreign_keys=[user_id])

        # Last date (date and user)
        creation_date = Column(UtcDateTime, nullable=False)

        # Usually it will be facility staff that will update this
        creation_user_id = Column(Integer, ForeignKey('users.id'),
                                     nullable=False)
        creation_user = relationship("User",
                                        foreign_keys=[creation_user_id])

        # Last date (date and user)
        last_update_date = Column(UtcDateTime, nullable=False)

        # Usually it will be facility staff that will update this
        last_update_user_id = Column(Integer, ForeignKey('users.id'),
                                     nullable=False)
        last_update_user = relationship("User",
                                        foreign_keys=[last_update_user_id])

        # General JSON dict to store extra attributes
        extra = Column(JSON, default={})

        entries = relationship('Entry', back_populates='project')

        def __getExtra(self, key, default):
            return self.extra.get(key, default)

        def __setExtra(self, key, value):
            extra = dict(self.extra or {})
            extra[key] = value
            self.extra = extra

        @property
        def user_can_edit(self):
            """ True if the user of the project can edit it (add/modify/delete notes)
            """
            return  self.__getExtra('user_can_edit', False)

        @user_can_edit.setter
        def user_can_edit(self, value):
            self.__setExtra('user_can_edit', value)

        def json(self):
            return dm.json_from_object(self)


    class Entry(Base):
        """ Entry related to a given project.
         """
        __tablename__ = 'entries'

        id = Column(Integer, primary_key=True)

        date = Column(UtcDateTime, nullable=False)

        type = Column(String(16), nullable=False)

        title = Column(String(256), nullable=False)

        description = Column(Text, nullable=True)

        # User to which this Project is associated with
        # The PI of this project
        project_id = Column(Integer, ForeignKey('projects.id'), nullable=False)
        project = relationship("Project", back_populates="entries")

        # Last date (date and user)
        creation_date = Column(UtcDateTime, nullable=False)

        # Usually it will be facility staff that will update this
        creation_user_id = Column(Integer, ForeignKey('users.id'),
                                     nullable=False)
        creation_user = relationship("User",
                                        foreign_keys=[creation_user_id])

        # Last date (date and user)
        last_update_date = Column(UtcDateTime, nullable=False)

        # Usually it will be facility staff that will update this
        last_update_user_id = Column(Integer, ForeignKey('users.id'),
                                     nullable=False)
        last_update_user = relationship("User",
                                        foreign_keys=[last_update_user_id])

        # General JSON dict to store extra attributes
        extra = Column(JSON, default={})

        def __getExtra(self, key, default):
            return self.extra.get(key, default)

        def __setExtra(self, key, value):
            extra = dict(self.extra or {})
            extra[key] = value
            self.extra = extra

        def json(self):
            return dm.json_from_object(self)


    class Puck(Base):
        """ Puck entity for Grids Storage table.
        """
        __tablename__ = 'pucks'

        id = Column(Integer,
                    primary_key=True)

        code = Column(String(64), unique=True, index=True)

        label = Column(String(64),
                      unique=True,
                      nullable=False)

        color = Column(String(16))

        # Locations properties
        dewar = Column(Integer)
        cane = Column(Integer)
        position = Column(Integer)

        # General JSON dict to store extra attributes
        extra = Column(JSON, default={})

        def json(self):
            return dm.json_from_object(self)

        def __getExtra(self, key, default):
            return self.extra.get(key, default)

        def __setExtra(self, key, value):
            extra = dict(self.extra or {})
            extra[key] = value
            self.extra = extra


    class PuckStorage:
        """ Simple class to organize pucks access. """

        def __init__(self, pucks):
            self._locDict = OrderedDict()
            self._idDict = OrderedDict()
            self._dewars = OrderedDict()

            def _locKey(p):
                return p.dewar * 10000 + p.cane * 100 + p.position

            last_dewar = 0
            last_cane = 0
            for puck in sorted(pucks, key=lambda p: _locKey(p)):
                puckJson = puck.json()
                d, c, p = loc = self.__puckLoc(puckJson)
                self._idDict[puck.id] = self._locDict[loc] = puckJson

                # Register dewar info (from first puck)
                if d != last_dewar:
                    dewar = dict(puck.extra.get('dewar', {}))
                    last_dewar = dewar['id'] = d
                    dewar['canes'] = OrderedDict()
                    self._dewars[d] = dewar

                if c != last_cane:
                    cane = dict(puck.extra.get('cane', {}))
                    last_cane = cane['id'] = c
                    dewar['canes'][c] = cane

        def __puckLoc(self, puck):
            return puck['dewar'], puck['cane'], puck['position']

        def __matchLoc(self, puck, loc):
            def _match(v1, v2):
                return v2 is None or v1 == v2

            pLoc = self.__puckLoc(puck)
            return all(_match(pLoc[i], loc[i]) for i in range(3))

        def pucks(self, dewar=None, cane=None, position=None):
            loc = (dewar, cane, position)

            for p in sorted(self._idDict.values(), key=lambda p: p['id']):
                if self.__matchLoc(p, loc):
                    yield p

        def get_puck(self, id_or_loc):
            """ Retrieve puck by id or by location. """
            d = self._locDict if isinstance(id_or_loc, tuple) else self._idDict
            return d[id_or_loc]

        def dewars(self):
            return self._dewars.values()

        def get_dewar(self, d):
            return self._dewars.get(d, None)

        def get_cane(self, d, c):
            cane = self._dewars[d]['canes'].get(c, None) if d in self._dewars else None
            return cane



    dm.Form = Form
    dm.User = User
    dm.Resource = Resource
    dm.Template = Template
    dm.Application = Application
    dm.Booking = Booking
    dm.Session = Session
    dm.Transaction = Transaction
    dm.InvoicePeriod = InvoicePeriod
    dm.Project = Project
    dm.Entry = Entry
    dm.Puck = Puck
    dm.PuckStorage = PuckStorage




import re


class Counter:
    HEADERS = ["", "Bookings", "Days", "%", "Cost"]
    FORMAT = u"{:>15}{:>10}{:>10}{:>10}{:>10}"
    TOTAL = None

    def __init__(self, name, filter=None):
        self._name = name
        self.counter = 0
        self.days = 0
        self.cost = 0
        self.bookings = []
        self._filter = filter or self._filter_by_name

    def _filter_by_name(self, b):
        return self._name.lower() in b['title'].lower()

    def count(self, b):
        if self._filter(b):
            self.counter += 1
            self.cost += b['total_cost']
            self.days += b['days']
            self.bookings.append(b)
            return True
        return False


class CounterList:
    def __init__(self, *names):
        self._counters = [
            Counter('Total', lambda b: True),
            Counter('Reminder', lambda b: True)
        ]
        self.reminder = []
        self._countersDict = {c._name: c for c in self._counters}

        for n in names:
            self.addCounter(n)


    def __getitem__(self, item):
        return self._countersDict[item]

    def addCounter(self, counter):
        c = counter if isinstance(counter, Counter) else Counter(str(counter))
        self._counters.insert(-1, c)
        self._countersDict[c._name] = c

    def count(self, b):
        self._counters[0].count(b)  # Always count total

        def _any():
            for c in self._counters[1:-1]:
                if c.count(b):
                    return True
            return False

        if not _any():
            self.reminder.append(b)
            self._counters[-1].count(b)  # Count reminder

    def data(self):
        total = self._counters[0].days
        data = [[c._name, c.counter, c.days,
                 '%0.2f' % (c.days * 100 / total), c.cost]
                for c in self._counters]

        return data

    def print(self):
        format = Counter.FORMAT.format
        print(format(*Counter.HEADERS))
        for row in self.data():
            print(format(*row))

    def printReminder(self):
        for b in self.reminder:
            print(b['title'])


MAINTENANCE_LIST = ['cycle', 'installation', 'maintenance', 'afis']
DEVELOPMENT_LIST = ['method', 'research', 'test', 'mikroed', 'microed', 'devel']

def _match_title(b, keywords):
    t = b['title'].lower()
    return any(k in t for k in keywords)

def is_maintenance(b):
    return b['type'] == 'maintenance' or _match_title(b, MAINTENANCE_LIST)


def is_development(b):
    return _match_title(b, DEVELOPMENT_LIST)


def get_cem(b):
    title = b['title'].upper()

    # Take only the first part of the application label
    # (because it can contains the alias in parenthesis)
    m = re.search("(CEM([0-9]+))", title)

    if m is not None:
        # Enforce numeric part is exactly 5 digits
        cemNumber = m.group(2)
        n = len(cemNumber)
        if n < 5:
            cemNumber = "0" * (5 - n) + cemNumber
        else:
            cemNumber = cemNumber[-5:]

        return 'CEM' + cemNumber

    return None


class CemCounter(Counter):
    def _filter_by_name(self, b):
        return self._name.upper() == get_cem(b)


def get_booking_counters(bookings):
    maintenance = Counter('Maintenance', is_maintenance)
    development = Counter('Development', is_development)
    CEM = Counter('CEM', filter=lambda b: get_cem(b) is not None)

    counters = CounterList('Downtime', maintenance, 'DBB', CEM, development)
    cem_counters = CounterList()

    cem_dict = {}

    for b in bookings:
        title = b['title']

        if 'Ume' in title:
            continue

        counters.count(b)

        cem = get_cem(b)

        if cem is not None:
            c = cem_dict.get(cem, 0)
            cem_dict[cem] = c + 1
            if not c:
                cem_counters.addCounter(CemCounter(cem))
            cem_counters.count(b)

    return counters, cem_counters






# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import unittest

from emhub.data import DataManager


class TestStringMethods(unittest.TestCase):
    print("=" * 80, "\nTesting string methods...")

    def setUp(cls):
        cls.ds = DataManager('/tmp/')

    def test_upper(self):
        self.assertEqual('foo'.upper(), 'FOO')

    def test_isupper(self):
        self.assertTrue('FOO'.isupper())
        self.assertFalse('Foo'.isupper())

    def test_split(self):
        s = 'hello world'
        self.assertEqual(s.split(), ['hello', 'world'])
        # check that s.split fails when the separator is not a string
        with self.assertRaises(TypeError):
            s.split(2)



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import os
import json
import requests
from contextlib import contextmanager


class config:
    EMHUB_SOURCE = os.environ['EMHUB_SOURCE']
    EMHUB_SERVER_URL = os.environ['EMHUB_SERVER_URL']
    EMHUB_USER = os.environ['EMHUB_USER']
    EMHUB_PASSWORD = os.environ['EMHUB_PASSWORD']


@contextmanager
def open_client():
    dc = DataClient(server_url=config.EMHUB_SERVER_URL)
    try:
        dc.login(config.EMHUB_USER, config.EMHUB_PASSWORD)
        yield dc
    finally:
        dc.logout()


class DataClient:
    """
    Simple client to communicate with the emhub REST API.
    """
    def __init__(self, server_url=None):
        self._server_url = server_url or os.environ.get('EMHUB_SERVER_URL',
                                                        'http://127.0.0.1:5000')
        # Store the last request object
        self.cookies = self.r = None

    def login(self, username=None, password=None):
        username = username or os.environ['EMHUB_USER']
        password = password or os.environ['EMHUB_PASSWORD']

        self.r = requests.post('%s/api/login' % self._server_url,
                               json={'username': username,
                                     'password': password})
        self.r.raise_for_status()
        self.cookies = self.r.cookies
        return self.r

    def logout(self):
        r = requests.post('%s/api/logout' % self._server_url,
                          cookies=self.cookies)
        r.raise_for_status()
        self.cookies = self.r = None
        return r

    def create_session(self, attrs):
        """ Request the server to create a new session.
        Mandatory in attrs:
            name: the session name
        """
        return self._method('create_session', 'session', attrs)

    def get_session(self, sessionId, attrs=None):
        return self._method('get_sessions', 0, attrs,
                            condition='id=%d' % sessionId)

    def update_session(self, attrs):
        """ Request the server to update existing session.
        Mandatory in attrs:
            id: the id of the session
        """
        return self._method('update_session', 'session', attrs)

    def delete_session(self, attrs):
        """ Request the server to delete a session.
        Mandatory in attrs:
            id: the id of the session
        """
        return self._method('delete_session', 'session', attrs)

    def create_session_set(self, attrs):
        """ Request the server to create a set within a session.
        Mandatory in attrs:
            session_id: the id of the session
            set_id: the id of the set that will be created
        """
        return self._method('create_session_set', 'session_set', attrs)

    def update_session_set(self, attrs):
        """ Request the server to update a set within a session.
        Mandatory in attrs:
            session_id: the id of the session
            set_id: the id of the set that will be created
        """
        return self._method('update_session_set', 'session_set', attrs)

    def get_session_sets(self, attrs):
        """ Retrieve all sets associated to a given session.
        Mandatory in attrs:
            session_id: the id of the session
        """
        return self._method('get_session_sets', 'session_sets', attrs)

    def add_session_item(self, attrs):
        """ Add new item to a set in the session.
        Mandatory in attrs:
            session_id: the id of the session
            set_id: the id of the set that will be created
            item_id: the id of the item to be added
        """
        return self._method('add_session_item', 'item', attrs)

    def update_session_item(self, attrs):
        """ Update existing item in the set in the session.
        Mandatory in attrs:
            session_id: the id of the session
            set_id: the id of the set
            item_id: the id of the item to be modified
        """
        return self._method('update_session_item', 'item', attrs)

    #---------------------- Internal functions ------------------------------
    def _method(self, method, resultKey, attrs, condition=None):
        r = self.request(method,
                         jsonData={'attrs': attrs,
                                   'condition': condition})
        json = r.json()
        if 'error' in json:
            raise Exception("ERROR from Server: ", json['error'])

        return json if resultKey is None else json[resultKey]

    def request(self, method, jsonData=None, bp='api'):
        """ Make a request to this method passing the json data.
        """
        if self.cookies is None:
            raise Exception("You should call login method first")

        self.r = requests.post('%s/%s/%s'
                               % (self._server_url, bp, method),
                               json=jsonData or {},  cookies=self.cookies)
        self.r.raise_for_status()
        return self.r

    def get(self, name, condition=None, orderBy=None, attrs=None):
        return self.request('get_%s' % name,
                            jsonData={'condition': condition,
                                      'orderBy': orderBy,
                                      'attrs': attrs})

    def json(self):
        if self.r.status_code == 200:
            return json.dumps(self.r.json(), indent=4)
        else:
            return {'ERROR': "Request failed with status code: %s"
                             % self.r.status_code}



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import os
import sys
import json
import requests

from emhub.utils import datetime_from_isoformat
from emhub.data import DataManager
from emhub.data.imports import TestDataBase


class PortalData(TestDataBase):
    """ Class to import data (users, templates, applications) from the
     Application Portal at SciLifeLab.
    """

    def __init__(self, dm, dataJsonPath, bookingsJsonPath):
        """
        Args:
            dm: DataManager with db to create test data
            dataJsonPath: file path to the JSON file with the data from the Portal.
                It is expected as a Dict with the following entries:
                * users
                * forms (Templates here)
                * orders (Applications here)
        """
        dm.create_admin()
        self.__importData(dm, dataJsonPath, bookingsJsonPath)

    def __importData(self, dm, dataJsonPath, bookingsJsonPath):
        print("Populating forms...")
        self._populateForms(dm)

        # Create tables with test data for each database model
        print("Importing users...")

        with open(dataJsonPath) as jsonFile:
            jsonData = json.load(jsonFile)
            self._jsonUsers = jsonData['users']['items']
            self._dictUsers = {}
            self.__importUsers(dm)

        print("Populating resources...")
        self._populateResources(dm)

        print("Importing applications...")
        self.__importApplications(dm, jsonData)

        with open(bookingsJsonPath) as bookingsFile:
            bookingsData = json.load(bookingsFile)

            # print("Populating sessions...")
            # self._populateSessions(dm)
            print("Populating bookings")
            self.__importBookings(dm, bookingsData)

    def __importUsers(self, dm):
        # Create user table
        def ignoreUser(u):
            if 'per.kraulis' in u['email']:
                return True
            return False

        def createUser(u, **kwargs):
            if ignoreUser(u):
                return

            status = 'inactive' if u['status'] == 'disabled' else 'active'

            roles = kwargs.get('roles', ['user'])
            pi = None
            if u['pi']:
                roles.append('pi')
            else:
                pi = kwargs.get('pi', None)

            user = dm.create_user(
                username=u['email'],
                email=u['email'],
                phone='',
                password=u['email'],
                name="%(first_name)s %(last_name)s" % u,
                roles=roles,
                pi_id=pi,
                status=status
            )

            u['emhub_item'] = user
            self._dictUsers[user.email] = user

        staff = {
            'marta.carroni@scilifelab.se': ['manager', 'head'],
            'julian.conrad@scilifelab.se': ['manager'],
            'karin.wallden@scilifelab.se': ['manager'],
            'mathieu.coincon@scilifelab.se': ['manager'],
            'dustin.morado@scilifelab.se': ['admin', 'manager'],
            'stefan.fleischmann@scilifelab.se': ['admin'],
            'delarosatrevin@scilifelab.se': ['admin'],
            # Umeå staff
            'michael.hall@umu.se': ['manager'],
            'camilla.holmlund@umu.se': ['manager'],
            'hussein.haggag@umu.se': ['admin']
        }

        #  Create first facility staff
        for u in self._jsonUsers:
            if u['email'] in staff:
                createUser(u, roles=staff[u['email']])

        # Insert first PI users, so we store their Ids for other users
        piDict = {}
        for u in self._jsonUsers:
            if u['pi']:
                createUser(u)
                piDict[u['email']] = u

        # f = open('users-missing-PI.csv', 'w')

        for u in self._jsonUsers:
            if not ignoreUser(u) and not u['pi'] and not u['email'] in staff:
                piEmail = u['invoice_ref']
                if piEmail in piDict:
                    createUser(u, pi=piDict[piEmail]['emhub_item'].id)
                else:
                    print("Skipping user (Missing PI): ", u['email'])
        #             f.write('"%s", \t"%s", \t"%s"\n'
        #                     % (u['name'], u['email'], piEmail))
        # f.close()

    def __importApplications(self, dm, jsonData):
        statuses = {'disabled': 'closed',
                    'review': 'review',
                    'preparation': 'preparation',
                    'enabled': 'active',
                    'rejected': 'rejected'
                    }

        def createTemplate(f):
            return dm.create_template(
                title=f['title'],
                description=f['description'],
                status=statuses[f['status']]
            )

        formsDict = {}

        internalTemplate = dm.create_template(
            title='Template for internal applications (DBB or FAC)',
            description='Special template for internal applications',
            status='closed')

        for f in jsonData['forms'].values():
            f['emhub_item'] = createTemplate(f)
            formsDict[f['iuid']] = f

        now = dm.now()

        def _internalPi(u):
            return (u['pi'] and 'emhub_item' in u and
                    (u['email'].endswith('dbb.su.se')
                     or u['email'].endswith('scilifelab.se')
                     or u['email'].endswith('mmk.su.se')))

        # Insert first PI users, so we store their Ids for other users
        dbbPis = [u['emhub_item'] for u in self._jsonUsers if _internalPi(u)]

        gunnar = self._dictUsers['gunnar@dbb.su.se']

        # Create special bag for internal DBB PIs
        internalApp = dm.create_application(
            code='DBB',
            title='Internal DBB Bag',
            created=now,  # datetime_from_isoformat(o['created']),
            alias='DBB',
            status='active',
            description='Internal application for DBB users',
            creator_id=gunnar.id,
            template_id=internalTemplate.id,
            invoice_reference='DBB invoice',
            resource_allocation={'quota': {'krios': 0, 'talos': 0},
                                 'noslot': [1, 2, 3]}  # allow to book scopes
        )

        for pi in dbbPis:
            internalApp.users.append(pi)

        import_date = now.replace(year=2018, month=6, day=1)

        def _alias(email):
            """ Try to guess the Application alias from PI email. """
            if email.endswith('kth.se'):
                return 'KTH Bag'
            if email.endswith('ki.se'):
                return 'KI Bag'
            if email.endswith('uu.se'):
                return 'Uppsala Bag'
            if email.endswith('umu.se'):
                return 'Umeå Bag'
            if email.endswith('gu.se'):
                return 'Gotherborg Bag'
            if email.endswith('.fi'):
                return 'Finland Bag'
            if email.endswith('.no'):
                return 'Norway Bag'

            return ''

        for o in jsonData['orders']:
            piEmail = o['owner']['email']
            orderId = o['identifier']

            pi = self._dictUsers.get(piEmail, None)

            if pi is None:
                print("Ignoring ORDER '%s', owner email (%s) not found as PI"
                      % (orderId, piEmail))
                continue

            status = o['status']
            # Set some accepted as 'active' and other as 'closed'
            # created = dt.datetime.strptime(o['created'], '%Y-%m-%d')
            created = datetime_from_isoformat(o['created'])

            # Skip too old applications
            if import_date > created:
                continue

            if status == 'accepted' or status == 'enabled':
                if created.year == now.year:
                    status = 'active'
                else:
                    status = 'closed'
            else:
                status = statuses[status]

            fields = o['fields']
            description = fields.get('project_des', None)
            invoiceRef = fields.get('project_invoice_addess', None)

            try:
                pi_list = fields.get('pi_list', [])
                alias = _alias(pi.email) if pi_list else ''

                app = dm.create_application(
                    code=orderId,
                    title=o['title'],
                    created=created,  # datetime_from_isoformat(o['created']),
                    alias=alias,
                    status=status,
                    description=description,
                    creator_id=pi.id,
                    template_id=formsDict[o['form']['iuid']]['emhub_item'].id,
                    invoice_reference=invoiceRef or 'MISSING_INVOICE_REF',
                )

                for piTuple in pi_list:
                    piEmail = piTuple[1]
                    pi = self._dictUsers.get(piEmail, None)
                    if pi is not None:
                        app.users.append(pi)

            except Exception as e:
                print("Exception when creating Application: %s. IGNORING..." % e)

        dm.commit()

    def __importBookings(self, dm, bookingsJson):
        now = dm.now().replace(minute=0, second=0)
        month = now.month

        resourcesDict = {
            'Titan Krios': 1,
            'Talos Arctica': 3,
            'Vitrobot': 4,
            'Carbon Coater': 6
        }

        # f = open('users-missing-PORTAL.csv', 'w')
        missing = set()

        for b in bookingsJson:
            name = b['user']['name']
            email = b['user']['email']
            resource = b['resourceName']

            if email not in self._dictUsers:
                if not email in missing:
                    missing.add(email)
                    # f.write('"%s", \t"%s"\n' % (name, email))

            if email not in self._dictUsers or resource not in resourcesDict:
                print(b['startDate'], b['endDate'], b['resourceName'],
                      b['title'], name)
                continue

            title = b['title']
            titleLow = title.lower()

            type = 'booking'
            if 'downtime' in titleLow:
                type = 'downtime'

            user = self._dictUsers[email]

            try:
                dm.create_booking(
                    check_min_booking=False,
                    check_max_booking=False,
                    title=b['title'],
                    start=datetime_from_isoformat(b['startDate']),
                    end=datetime_from_isoformat(b['endDate']),
                    type=type,
                    resource_id=resourcesDict[resource],
                    creator_id=user.id,  # first user for now
                    owner_id=user.id,  # first user for now
                    description="")
            except Exception as e:
                print("Exception when creating Booking: %s. IGNORING..." % e)

        # f.close()

        # create a repeating event
        dm.create_booking(title='Dropin',
                          start=now.replace(day=6, hour=9),
                          end=now.replace(day=6, hour=13),
                          type='slot',
                          repeat_value='bi-weekly',
                          repeat_stop=now.replace(month=month + 2),
                          resource_id=7,
                          creator_id=1,  # first user for now
                          owner_id=1,  # first user for now
                          description="Recurrent bi-weekly DROPIN slot. ")

    def _populateSessions(self, dm):
        users = [1, 2, 2]
        session_names = ['supervisor_23423452_20201223_123445',
                         'epu-mysession_20122310_234542',
                         'mysession_very_long_name']

        testData = os.environ.get('EMHUB_TESTDATA')
        fns = [os.path.join(testData, 'hdf5/20181108_relion30_tutorial.h5'),
               os.path.join(testData, 'hdf5/t20s_pngs.h5'), 'non-existing-file']

        scopes = ['Krios 1', 'Krios 2', 'Krios 3']
        numMovies = [423, 234, 2543]
        numMics = [0, 234, 2543]
        numCtfs = [0, 234, 2543]
        numPtcls = [0, 2, 2352534]
        status = ['Running', 'Error', 'Finished']

        for f, u, s, st, sc, movies, mics, ctfs, ptcls in zip(fns, users, session_names,
                                                              status, scopes, numMovies,
                                                              numMics, numCtfs, numPtcls):
            dm.create_session(
                sessionData=f,
                userid=u,
                sessionName=s,
                dateStarted=dm.now(),
                description='Long description goes here.....',
                status=st,
                microscope=sc,
                voltage=300,
                cs=2.7,
                phasePlate=False,
                detector='Falcon',
                detectorMode='Linear',
                pixelSize=1.1,
                dosePerFrame=1.0,
                totalDose=35.0,
                exposureTime=1.2,
                numOfFrames=48,
                numOfMovies=movies,
                numOfMics=mics,
                numOfCtfs=ctfs,
                numOfPtcls=ptcls,
                numOfCls2D=0,
                ptclSizeMin=140,
                ptclSizeMax=160,
            )

        dm.create_session(sessionData='dfhgrth',
                          userid=2,
                          sessionName='dfgerhsrth_NAME',
                          dateStarted=dm.now(),
                          description='Long description goes here.....',
                          status='Running',
                          microscope='KriosX',
                          voltage=300,
                          cs=2.7,
                          phasePlate=False,
                          detector='Falcon',
                          detectorMode='Linear',
                          pixelSize=1.1,
                          dosePerFrame=1.0,
                          totalDose=35.0,
                          exposureTime=1.2,
                          numOfFrames=48,
                          numOfMovies=0,
                          numOfMics=0,
                          numOfCtfs=0,
                          numOfPtcls=0,
                          numOfCls2D=0,
                          ptclSizeMin=140,
                          ptclSizeMax=160, )


class PortalManager:
    """ Helper class to interact with the Application Portal system.
    """
    def __init__(self, apiJson, cache=True):
        self._headers = apiJson['headers']
        self._baseUrl = apiJson['baseUrl']

        # Create a cached dict with json files for url
        # to avoid make unnecessary queries in the same session
        if cache:
            self._cache = {}

    def _getUrl(self, suffix):
        return self._baseUrl + suffix

    def _fetchJsonFromUrl(self, url):
        cache = getattr(self, '_cache', None)
        cachedJson = cache.get(url, None) if cache is not None else None

        if cachedJson:
            print("Returning cached JSON for url: %s" % url)
            return cachedJson

        print("Retrieving url: %s" % url)
        response = requests.get(url, headers=self._headers)

        if response.status_code != 200:
            print(response.status_code)
            return None
        else:
            result = response.json()
            if cache is not None:
                cache[url] = result
            return result

    def _fetchJsonFromUrlSuffix(self, suffix):
        return self._fetchJsonFromUrl(self._getUrl(suffix))

    def fetchOrdersJson(self):
        """ Fetch orders from the Portal. """
        return self._fetchJsonFromUrlSuffix('orders?recent=False')['items']

    def fetchOrderDetailsJson(self, orderCEM):
        return self._fetchJsonFromUrlSuffix('order/%s' % orderCEM.upper())
        #orderUrl = orderJson['links']['api']['href']
        #return self._fetchJsonFromUrl(orderUrl)

    def fetchFormDetailsJson(self, formId):
        """ Return  the JSON data about a given form. """
        return self._fetchJsonFromUrlSuffix('form/%s' % formId)

    def fetchAccountsJson(self):
        """ Retrieve the users list from the portal system. """
        return self._fetchJsonFromUrlSuffix('accounts')['items']

    def fetchAccountDetailsJson(self, userEmail):
        return self._fetchJsonFromUrlSuffix('account/%s' % userEmail)


if __name__ == '__main__':
    instance_path = os.path.abspath(os.environ.get("EMHUB_INSTANCE",
                                                   'instance'))

    if not os.path.exists(instance_path):
        raise Exception("Instance folder '%s' not found!!!" % instance_path)

    if len(sys.argv) != 3:
        print("\nUSAGE:\n"
              "    python -m emhub.data.imports.scilifelab portal-data.json booked-data.json\n")
        sys.exit(1)

    portalDataJson = sys.argv[1]
    bookingsJson = sys.argv[2]

    if not os.path.exists(portalDataJson):
        print("JSON data file '%s' does not exists. " % portalDataJson)

    if not os.path.exists(bookingsJson):
        print("JSON bookings file '%s' does not exists. " % bookingsJson)

    dm = DataManager(instance_path, cleanDb=True)
    PortalData(dm, portalDataJson, bookingsJson)



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import datetime as dt
import os
import uuid
from collections import defaultdict

import sqlalchemy

from emhub.utils import datetime_from_isoformat, datetime_to_isoformat
from .data_db import DbManager
from .data_log import DataLog
from .data_models import create_data_models
from .data_session import H5SessionData


class DataManager(DbManager):
    """ Main class that will manage the sessions and their information.
    """
    def __init__(self, dataPath, dbName='emhub.sqlite',
                 user=None, cleanDb=False, create=True):
        self._dataPath = dataPath
        self._sessionsPath = os.path.join(dataPath, 'sessions')
        self._entryFiles = os.path.join(dataPath, 'entry_files')
        self._resourceFiles = os.path.join(dataPath, 'resource_files')

        # Initialize main database
        dbPath = os.path.join(dataPath, dbName)
        self.init_db(dbPath, cleanDb=cleanDb, create=create)

        self._lastSession = None
        self._user = user  # Logged user

        if create:
            # Create a separate database for logs
            logDbPath = dbPath.replace('.sqlite', '-logs.sqlite')
            self._db_log = DataLog(logDbPath, cleanDb=cleanDb)

            # Create sessions dir if not exists
            os.makedirs(self._sessionsPath, exist_ok=True)

    def _create_models(self):
        """ Function called from the init_db method. """
        create_data_models(self)

    def log(self, log_type, log_name, *args, **kwargs):
        log_user_id = None if self._user is None else self._user.id

        self._db_log.log(log_user_id, log_type, log_name,
                         *args, **kwargs)

    def get_logs(self):
        return self._db_log.get_logs()

    # ------------------------- USERS ----------------------------------
    def create_admin(self, password='admin'):
        """ Create special user 'admin'. """
        admin = self.create_user(username='admin',
                                 email='admin@emhub.org',
                                 password=password,
                                 name='admin',
                                 roles=['admin'],
                                 pi_id=None)
        if self._user is None:
            self._user = admin

    def create_user(self, **attrs):
        """ Create a new user in the DB. """
        attrs['password_hash'] = self.User.create_password_hash(attrs['password'])
        del attrs['password']
        return self.__create_item(self.User, **attrs)

    def update_user(self, **attrs):
        """ Update an existing user. """
        if 'password' in attrs:
            attrs['password_hash'] = self.User.create_password_hash(attrs['password'])
            del attrs['password']

        return self.__update_item(self.User, **attrs)

    def delete_user(self, **attrs):
        """ Delete a given user. """
        user = self.__item_by(self.User, id=attrs['id'])
        self.delete(user)
        return user

    def get_users(self, condition=None, orderBy=None, asJson=False):
        return self.__items_from_query(self.User,
                                       condition=condition,
                                       orderBy=orderBy,
                                       asJson=asJson)

    def get_user_by(self, **kwargs):
        """ This should return a single user or None. """
        return self.__item_by(self.User, **kwargs)

    # ---------------------------- FORMS ---------------------------------
    def create_form(self, **attrs):
        return self.__create_item(self.Form, **attrs)

    def update_form(self, **attrs):
        return self.__update_item(self.Form, **attrs)

    def delete_form(self, **attrs):
        form = self.__item_by(self.Form, id=attrs['id'])
        self.delete(form)
        return form

    def get_forms(self, condition=None, orderBy=None, asJson=False):
        return self.__items_from_query(self.Form,
                                       condition=condition,
                                       orderBy=orderBy,
                                       asJson=asJson)

    def get_form_by(self, **kwargs):
        """ This should return a single Form or None. """
        return self.__item_by(self.Form, **kwargs)

    def get_form_by_name(self, formName):
        """ Shortcut method to load a form from db given its name.
        If the form does not exist, an Exception is thrown.
        """
        form = self.get_form_by(name=formName)
        if form is None:
            raise Exception("Missing Form '%s' from the database!!!" % formName)

        return form

    # ---------------------------- RESOURCES ---------------------------------
    def create_resource(self, **attrs):
        return self.__create_item(self.Resource, **attrs)

    def update_resource(self, **attrs):
        return self.__update_item(self.Resource, **attrs)

    def get_resources(self, condition=None, orderBy=None, asJson=False):
        return self.__items_from_query(self.Resource,
                                       condition=condition,
                                       orderBy=orderBy,
                                       asJson=asJson)

    def get_resource_by(self, **kwargs):
        """ This should return a single Resource or None. """
        return self.__item_by(self.Resource, **kwargs)

    def delete_resource(self, **attrs):
        resource = self.__item_by(self.Resource, id=attrs['id'])
        self.delete(resource)
        return resource

    def get_resource_image_path(self, resource, filename=None):
        fn = filename or resource.image
        return os.path.join(self._resourceFiles,
                            'resource-image-%06d-%s' % (resource.id, fn))

    # ---------------------------- APPLICATIONS --------------------------------
    def create_template(self, **attrs):
        return self.__create_item(self.Template, **attrs)

    def get_templates(self, condition=None, orderBy=None, asJson=False):
        return self.__items_from_query(self.Template,
                                       condition=condition,
                                       orderBy=orderBy,
                                       asJson=asJson)

    def get_template_by(self, **kwargs):
        """ Return a single Template or None. """
        return self.__item_by(self.Template, **kwargs)

    def update_template(self, **attrs):
        return self.__update_item(self.Template, **attrs)

    def delete_template(self, **attrs):
        template = self.__item_by(self.Template, id=attrs['id'])
        self.delete(template)
        return template

    def create_application(self, **attrs):
        def __create(attrs):
            pi_to_add = attrs.pop('pi_to_add', [])
            pi_to_remove = attrs.pop('pi_to_remove', [])
            application = self.Application(**attrs)
            self.__update_application_pi(application, pi_to_add, pi_to_remove)
            return application

        attrs['special_create'] = __create
        attrs['created'] = self.now()
        attrs['invoice_reference'] = ''  #FIXME
        if 'creator' not in attrs:
            attrs['creator_id'] = self._user.id
        return self.__create_item(self.Application, **attrs)

    def get_applications(self, condition=None, orderBy=None, asJson=False):
        return self.__items_from_query(self.Application,
                                       condition=condition,
                                       orderBy=orderBy,
                                       asJson=asJson)

    def get_visible_applications(self):
        return [a for a in self.get_applications()
                if a.allows_access(self._user)]

    def get_application_by(self, **kwargs):
        """ Return a single Application or None. """
        return self.__item_by(self.Application, **kwargs)

    def delete_application(self, **attrs):
        application = self.__item_by(self.Application, id=attrs['id'])
        self.delete(application)
        return application

    def __update_application_pi(self, application, pi_to_add, pi_to_remove):
        errorMsg = ""
        pi_list = application.pi_list

        def _get_pi(pid):
            pi = self.get_user_by(id=int(pid))

            if pi is None:
                errorMsg += "\nInvalid user id: %s" % pid
            elif not pi.is_pi:
                errorMsg += "\nUser %s is not " % pid
            else:
                return pi
            return None

        to_add = []
        for pid in pi_to_add:
            pi = _get_pi(pid)

            if pi is None:
                continue

            if pi in pi_list:
                errorMsg += "\nPI %s is already in the Application" % pi.name
                continue

            to_add.append(pi)

        to_remove = []
        for pid in pi_to_remove:
            pi = _get_pi(pid)

            if pi is None:
                continue

            if pi not in pi_list:
                errorMsg += "\nPI %s is not in the Application" % pi.name
                continue

            to_remove.append(pi)

        if errorMsg:
            raise Exception(errorMsg)

        for pi in to_remove:
            application.users.remove(pi)

        for pi in to_add:
            application.users.append(pi)

    def __preprocess_application(self, application, attrs):
        pi_to_add = attrs.pop('pi_to_add', [])
        pi_to_remove = attrs.pop('pi_to_remove', [])
        self.__update_application_pi(application, pi_to_add, pi_to_remove)

    def update_application(self, **attrs):
        """ Update a given Application with new attributes.
        Special case are:
            pi_to_add: ids of PI users to add to the Application.
            pi_to_remove: ids of PI users to remove from the Application
        """
        attrs['special_update'] = self.__preprocess_application
        return self.__update_item(self.Application, **attrs)

    # ---------------------------- BOOKINGS -----------------------------------
    def create_booking(self,
                       check_min_booking=True,
                       check_max_booking=True,
                       **attrs):
        # We might create many bookings if repeat != 'no'
        repeat_value = attrs.get('repeat_value', 'no')
        modify_all = attrs.pop('modify_all', None)
        bookings = []

        def _add_booking(attrs):
            b = self.__create_booking(attrs,
                                      check_min_booking=check_min_booking,
                                      check_max_booking=check_max_booking)
            bookings.append(b)

        if repeat_value == 'no':
            _add_booking(attrs)
        else:
            repeat_stop = attrs.pop('repeat_stop')
            repeater = RepeatRanges(repeat_value, attrs)
            uid = str(uuid.uuid4())

            while attrs['end'] < repeat_stop:
                attrs['repeat_id'] = uid
                _add_booking(attrs)
                repeater.move()  # will move next start,end in attrs

        # Insert all created bookings
        for b in bookings:
            self._db_session.add(b)
        self.commit()

        # Log operations after create
        self.log('operation', 'create_Booking',
                 check_min_booking=check_min_booking,
                 check_max_booking=check_max_booking,
                 modify_all=modify_all,
                 repeat_stop=repeat_value,
                 attrs=self.json_from_dict(attrs))

        return bookings

    def update_booking(self, **attrs):
        """ Update one or many bookings (in case of repeating events)

        Keyword Args:
            id: the of the booking to be updated
            modify_all: Boolean flag in case the booking is a repeating event.
                If True, all bookings from this one, will be also updated.
        """
        repeat = attrs.get('repeat_value', 'no')
        repeater = RepeatRanges(repeat, attrs) if repeat != 'no' else None

        def update(b):
            self.__check_cancellation(b, attrs)

            for attr, value in attrs.items():
                if attr != 'id':
                    setattr(b, attr, value)
            if repeater:
                repeater.move()  # move start, end for repeating bookings

            self.__validate_booking(b)

        result = self._modify_bookings(attrs, update)

        self.log('operation', 'update_Booking',
                 attrs=self.json_from_dict(attrs))

        return result

    def get_bookings(self, condition=None, orderBy=None, asJson=False):
        return self.__items_from_query(self.Booking,
                                       condition=condition,
                                       orderBy=orderBy,
                                       asJson=asJson)

    def get_bookings_range(self, start, end, resource=None):
        """ Shortcut function to retrieve a range of bookings. """
        # JMRT: For some reason the retrieval of the date ranges is not working
        # as expected for the time. So we are taking on day before for the start
        # and one day after for the end and filter later
        newStart = (start - dt.timedelta(days=1)).replace(hour=23, minute=59)
        newEnd = (end + dt.timedelta(days=1)).replace(hour=0, minute=0)
        rangeStr = datetime_to_isoformat(newStart), datetime_to_isoformat(newEnd)

        startBetween = "(start>='%s' AND start<='%s')" % rangeStr
        endBetween = "(end>='%s' AND end<='%s')" % rangeStr
        rangeOver = "(start<='%s' AND end>='%s')" % rangeStr
        conditionStr = "(%s OR %s OR %s)" % (startBetween, endBetween, rangeOver)

        if resource is not None:
            conditionStr += " AND resource_id=%s" % resource.id

        def in_range(b):
            s, e = b.start, b.end
            return ((s >= start and s <= end) or
                    (e >= start and e <= end) or
                    (s <= start and e >= end))

        return [b for b in self.get_bookings(condition=conditionStr, orderBy='start')
                if in_range(b)]

    def get_next_bookings(self, user):
        """ Retrieve upcoming (from now) bookings for this user. """
        conditionStr = "start>='%s'" % datetime_to_isoformat(self.now())
        if user:
            conditionStr += " AND owner_id=%s" % user.id

        return self.get_bookings(condition=conditionStr, orderBy='start')

    def delete_booking(self, **attrs):
        """ Delete one or many bookings (in case of repeating events)

        Keyword Args:
            id: the of the booking to be deleted
            modify_all: Boolean flag in case the booking is a repeating event.
                If True, all bookings from this one, will be also deleted.
        """
        def delete(b):
            self.__check_cancellation(b)
            self.delete(b, commit=False)

        result = self._modify_bookings(attrs, delete)

        self.log('operation', 'delete_Booking',
                 attrs=self.json_from_dict(attrs))

        return result

    def get_application_bookings(self, applications,
                                resource_ids=None, resource_tags=None):
        pass

    def count_booking_resources(self, applications,
                                resource_ids=None, resource_tags=None):
        """ Count how many days has been used by applications from the
        current bookings. The count can be done by resources or by tags.
        """
        application_ids = set(a for a in applications)
        count_dict = defaultdict(lambda: defaultdict(lambda: 0))

        for b in self.get_bookings():
            if b.application is None:
                continue

            baid = b.application.id
            if baid in application_ids:
                rid = b.resource.id
                if resource_tags is not None:
                    for tag in resource_tags:
                        if tag in b.resource.tags:
                            count_dict[baid][tag] += b.days
                elif not resource_ids or rid in resource_ids:
                    count_dict[baid][rid] += b.days

        return count_dict

    # ---------------------------- SESSIONS -----------------------------------
    def __get_section(self, sectionName):
        formDef = self.get_form_by_name('sessions_config').definition
        for s in formDef['sections']:
            if s['label'] == sectionName:
                return formDef, s
        return None

    def __iter_config_params(self, configName):
        _, section = self.__get_section(configName)
        for p in section['params']:
            yield p

    def __get_session_dict(self, section):
        return {p['label']: p['value']
                for p in self.__iter_config_params(section)}

    def get_session_counter(self, group_code):
        return int(self.__get_session_dict('counters').get(group_code, 1))

    def update_session_counter(self, group_code, new_counter):
        # Update counter for this session group
        formDef, section = self.__get_section('counters')

        found = False
        for p in section['params']:
            if p['label'] == group_code:
                p['value'] = new_counter
                found = True
        if not found:
            section['params'].append({'label': group_code,
                                      'value': new_counter})

        form = self.get_form_by_name('sessions_config')
        self.update_form(id=form.id, definition=formDef)

    def get_session_cameras(self, resourceId):
        cameras = []
        for p in self.__iter_config_params('cameras'):
            if int(p['id']) == resourceId:
                cameras = p['enum']['choices']

        return cameras

    def get_session_folders(self):
        return self.__get_session_dict('folders')

    def get_session_data_deletion(self, group_code):
        return int(self.__get_session_dict('data_deletion')[group_code])

    def get_session_processing(self):
        # Load processing options from the 'processing' Form
        form_proc = self.get_form_by_name('processing')

        processing = []
        for section in form_proc.definition['sections']:
            steps = []
            processing.append({'name': section['label'], 'steps': steps})
            for param in section['params']:
                steps.append({'name': param['label'], 'options': param['enum']['choices']})

        return processing

    def get_session_data_path(self, session):
        return os.path.join(self._sessionsPath, session.data_path)

    def get_new_session_info(self, booking_id):
        """ Return the name for the new session, base on the booking and
        the previous sessions counter (stored in Form 'counters').
        """
        b = self.get_bookings(condition="id=%s" % booking_id)[0]
        a = b.application
        code = 'fac' if a is None else a.code.lower()
        sep = '' if len(code) == 3 else '_'
        c = self.get_session_counter(code)

        return {
            'code': code,
            'counter': c,
            'name': '%s%s%05d' % (code, sep, c)
        }

    def get_sessions(self, condition=None, orderBy=None, asJson=False):
        """ Returns a list.
        condition example: text("id<:value and name=:name")
        """
        return self.__items_from_query(self.Session,
                                       condition=condition,
                                       orderBy=orderBy,
                                       asJson=asJson)

    def get_session_by(self, **kwargs):
        """ This should return a single Session or None. """
        return self.__item_by(self.Session, **kwargs)

    def create_session(self, **attrs):
        """ Add a new session row. """
        create_data = attrs.pop('create_data', False)
        b = self.get_bookings(condition="id=%s" % attrs['booking_id'])[0]
        attrs['resource_id'] = b.resource.id
        attrs['operator_id'] = b.owner.id if b.operator is None else b.operator.id

        if 'start' not in attrs:
            attrs['start'] = self.now()

        if 'status' not in attrs:
            attrs['status'] = 'pending'

        session_info = self.get_new_session_info(b.id)
        attrs['name'] = session_info['name']

        session = self.__create_item(self.Session, **attrs)

        # Let's update the data path after we know the id
        session.data_path = 'session_%06d.h5' % session.id

        self.commit()

        # Create empty hdf5 file
        if create_data:
            data = H5SessionData(self._session_data_path(session), mode='a')
            data.close()

        # Update counter for this session group
        self.update_session_counter(session_info['code'],
                                    session_info['counter'] + 1)

        return session

    def update_session(self, **attrs):
        """ Update session attrs. """
        session = self.__update_item(self.Session, **attrs)

        # Update the session counter if it was modified
        name = session.name

        if '_' in name:
            code, counterStr = name.split('_')
        else:
            code = name[:3]
            counterStr = name[3:]

        c = self.get_session_counter(code)
        counter = int(counterStr)
        if c < counter:
            self.update_session_counter(code, counter + 1)

        return session

    def delete_session(self, **attrs):
        """ Remove a session row. """
        sessionId = attrs['id']
        session = self.Session.query.get(sessionId)
        data_path = self._session_data_path(session)
        self.delete(session)

        if os.path.exists(data_path):
            os.remove(data_path)

        self.log("operation", "delete_Session",
                 attrs=self.json_from_dict(attrs))

        return session

    def load_session(self, sessionId, mode="r"):
        # if self._lastSession is not None:
        #     if self._lastSession.id == sessionId:
        #         return self._lastSession
        #     self._lastSession.data.close()

        session = self.Session.query.get(sessionId)
        session.data = H5SessionData(self._session_data_path(session), mode)
        return session

    def clear_session_data(self, **attrs):
        session = self.get_session_by(id=attrs['id'])
        data_path = self._session_data_path(session)
        if os.path.exists(data_path):
            os.remove(data_path)
        return session

    # -------------------------- INVOICE PERIODS ------------------------------
    def get_invoice_periods(self, condition=None, orderBy=None, asJson=False):
        """ Returns a list.
        condition example: text("id<:value and name=:name")
        """
        return self.__items_from_query(self.InvoicePeriod,
                                       condition=condition,
                                       orderBy=orderBy,
                                       asJson=asJson)

    def create_invoice_period(self, **attrs):
        """ Add a new session row. """
        return self.__create_item(self.InvoicePeriod, **attrs)

    def update_invoice_period(self, **attrs):
        """ Update session attrs. """
        return self.__update_item(self.InvoicePeriod, **attrs)

    def delete_invoice_period(self, **attrs):
        """ Remove a session row. """
        return self.__delete_item(self.InvoicePeriod, **attrs)

    def get_invoice_period_by(self, **kwargs):
        """ This should return a single user or None. """
        return self.__item_by(self.InvoicePeriod, **kwargs)

    # ---------------------------- TRANSACTIONS -------------------------------
    def get_transactions(self, condition=None, orderBy=None, asJson=False):
        """ Returns a list.
        condition example: text("id<:value and name=:name")
        """
        return self.__items_from_query(self.Transaction,
                                       condition=condition,
                                       orderBy=orderBy,
                                       asJson=asJson)

    def create_transaction(self, **attrs):
        """ Add a new session row. """
        return self.__create_item(self.Transaction, **attrs)

    def update_transaction(self, **attrs):
        """ Update session attrs. """
        return self.__update_item(self.Transaction, **attrs)

    def delete_transaction(self, **attrs):
        """ Remove a session row. """
        return self.__delete_item(self.Transaction, **attrs)

    def get_transaction_by(self, **kwargs):
        """ This should return a single user or None. """
        return self.__item_by(self.Transaction, **kwargs)

    # ---------------------------- PROJECTS ---------------------------------
    def __check_project(self, **attrs):
        if 'title' in attrs:
            if not attrs['title'].strip():
                raise Exception("Project title can not be empty")

        if 'user_id' in attrs:
            if not attrs['user_id']:
                raise Exception("Provide a valid User ID for the Project.")

        if 'status' in attrs:
            if not attrs['status'].strip() in self.Project.STATUS:
                raise Exception("Provide a valid status: active/inactive")

    def create_project(self, **attrs):
        self.__check_project(**attrs)

        now = self.now()
        attrs.update({
            'date': now,
            'creation_date': now,
            'creation_user_id': self._user.id,
            'last_update_date': now,
            'last_update_user_id': self._user.id,
        })
        return self.__create_item(self.Project, **attrs)

    def update_project(self, **attrs):
        self.__check_project(**attrs)

        attrs.update({
            'last_update_date': self.now(),
            'last_update_user_id': self._user.id,
        })
        return self.__update_item(self.Project, **attrs)

    def delete_project(self, **attrs):
        """ Remove a session row. """
        project = self.get_project_by(id=attrs['id'])
        # Delete all entries of this project
        # (since I haven't configured cascade-delete in SqlAlchemy models)
        for e in project.entries:
            self.delete(e, commit=False)

        return self.__delete_item(self.Project, **attrs)

    def get_projects(self, condition=None, orderBy=None, asJson=False):
        return self.__items_from_query(self.Project,
                                       condition=condition,
                                       orderBy=orderBy,
                                       asJson=asJson)

    def get_project_by(self, **kwargs):
        """ This should return a single Resource or None. """
        return self.__item_by(self.Project, **kwargs)

    # ---------------------------- ENTRIES ---------------------------------
    def __check_entry(self, **attrs):
        if 'title' in attrs:
            if not attrs['title'].strip():
                raise Exception("Entry title can not be empty")

        # if 'type' in attrs:
        #     entry_types = self.get_entry_types()
        #     if not attrs['type'].strip() in entry_types:
        #         raise Exception("Please provide a valid entry type: %s"
        #                         % entry_types)

    def create_entry(self, **attrs):
        self.__check_entry(**attrs)

        now = self.now()
        attrs.update({
            'date': now,
            'creation_date': now,
            'creation_user_id': self._user.id,
            'last_update_date': now,
            'last_update_user_id': self._user.id,
        })
        return self.__create_item(self.Entry, **attrs)

    def update_entry(self, **attrs):
        self.__check_entry(**attrs)

        attrs.update({
            'last_update_date': self.now(),
            'last_update_user_id': self._user.id,
        })
        return self.__update_item(self.Entry, **attrs)

    def delete_entry(self, **attrs):
        """ Remove a session row. """
        return self.__delete_item(self.Entry, **attrs)

    def get_entries(self, condition=None, orderBy=None, asJson=False):
        return self.__items_from_query(self.Entry,
                                       condition=condition,
                                       orderBy=orderBy,
                                       asJson=asJson)

    def get_entry_by(self, **kwargs):
        """ This should return a single Resource or None. """
        return self.__item_by(self.Entry, **kwargs)

    def get_entry_path(self, entry, filename):
        return os.path.join(self._entryFiles,
                            'entry-file-%06d-%s' % (entry.id, filename))

    def get_entry_file(self, entry, file_key):
        """ Return the fn associated with a given entry. """
        fn = entry.extra['data'].get(file_key, None)

        if fn is None:
            raise Exception("Can not retrieve file for key '%s'" % file_key)

        return self.get_entry_path(entry, fn)

    def get_entry_files(self, entry):
        """ Return all values from the extra dict that are files. """
        data = entry.extra['data']

        def _is_file(k):
            if k.endswith('_file'):
                return True
            return False

        files = []

        def _add_from_dict(d):
            if not isinstance(d, dict):
                return

            for k, v in d.items():
                if isinstance(v, list):
                    for row in v:
                        _add_from_dict(row)
                elif _is_file(k):
                    files.append(self.get_entry_path(entry, v))

        _add_from_dict(data)
        return files

    # ---------------------------- PUCKS ---------------------------------
    def create_puck(self, **attrs):
        return self.__create_item(self.Puck, **attrs)

    def update_puck(self, **attrs):
        return self.__update_item(self.Puck, **attrs)

    def delete_puck(self, **attrs):
        return self.__delete_item(self.Puck, **attrs)

    def get_pucks(self, condition=None, orderBy=None, asJson=False):
        return self.__items_from_query(self.Puck,
                                       condition=condition,
                                       orderBy=orderBy,
                                       asJson=asJson)

    def get_puck_by(self, **kwargs):
        return self.__item_by(self.Entry, **kwargs)

    # --------------- Internal implementation methods -------------------------
    def get_universities_dict(self):
        formDef = self.get_form_by_name('universities').definition
        return {p['label']: p['value'] for p in formDef['params']}

    def __create_item(self, ModelClass, **attrs):
        special_create = attrs.pop('special_create', None)
        jsonArgs = self.json_from_dict(attrs)

        new_item = special_create(attrs) if special_create else ModelClass(**attrs)

        self._db_session.add(new_item)
        self.commit()
        self.log('operation', 'create_%s' % ModelClass.__name__, attrs=jsonArgs)

        return new_item

    def __items_from_query(self, ModelClass,
                           condition=None, orderBy=None, asJson=False):
        query = self._db_session.query(ModelClass)

        if condition is not None:
            query = query.filter(sqlalchemy.text(condition))

        if orderBy is not None:
            query = query.order_by(orderBy)

        result = query.all()
        return [s.json() for s in result] if asJson else result

    def __item_by(self, ModelClass, **kwargs):
        query = self._db_session.query(ModelClass)
        return query.filter_by(**kwargs).one_or_none()

    def __update_item(self, ModelClass, **kwargs):
        special_update = kwargs.pop('special_update', None)

        jsonArgs = self.json_from_dict(kwargs)

        item_id = kwargs.pop('id')
        extra_replace = kwargs.pop('extra_replace', False)

        item = self.__item_by(ModelClass, id=item_id)
        if item is None:
            raise Exception("Not found item %s with id %s"
                            % (ModelClass.__name__, kwargs['id']))

        if special_update:
            special_update(item, kwargs)

        for attr, value in kwargs.items():
            if attr == 'extra' and not extra_replace:
                if not extra_replace:
                    extra = dict(item.extra)
                    extra.update(value)
                    value = extra
            setattr(item, attr, value)
        self.commit()
        self.log('operation', 'update_%s' % ModelClass.__name__, attrs=jsonArgs)

        return item

    def __delete_item(self, ModelClass, **kwargs):
        """ Remove an item from a Db model table. """
        item = self.__item_by(ModelClass, id=kwargs['id'])
        self.delete(item)

        self.log("operation", "delete_%s" % ModelClass.__name__,
                 attrs=self.json_from_dict(kwargs))

        return item

    def __matching_project(self, applications, auth_json):
        """ Return True if there is a project code in auth_json. """
        if not auth_json or 'applications' not in auth_json:
            return False

        json_codes = auth_json['applications']

        return any(p.code in json_codes for p in applications)

    # ------------------- BOOKING helper functions -----------------------------
    def __create_booking(self, attrs, **kwargs):
        if 'creator_id' not in attrs:
            attrs['creator_id'] = self._user.id

        b = self.Booking(**attrs)
        self.__validate_booking(b, **kwargs)
        return b

    def __validate_booking(self, booking, **kwargs):
        r = self.get_resource_by(id=booking.resource_id)
        check_min_booking = kwargs.get('check_min_booking', True)
        check_max_booking = kwargs.get('check_max_booking', True)

        # Booking starting date should be before ending date
        if booking.start >= booking.end:
            raise Exception("The booking 'end' should be after the 'start'. ")

        user = self._user

        # The following validations do not apply for managers
        if not user.is_manager:
            # Selected resource should be active
            if not r.is_active:
                raise Exception("Selected resource is inactive now. ")

            # Booking can not be made in the past
            if booking.start.date() < self.now().date():
                raise Exception("The booking 'start' can not be in the past. ")

            # Booking time should be bigger than the minimum for this resource
            if check_min_booking and r.min_booking > 0:
                mm = dt.timedelta(minutes=int(r.min_booking * 60))
                if booking.duration < mm:
                    raise Exception("The duration of the booking is less that "
                                    "the minimum specified for the resource. ")

            # Booking time should be less than the maximum for this resource
            if booking.type == 'booking' and check_max_booking and r.max_booking > 0:
                mm = dt.timedelta(minutes=int(r.max_booking * 60))
                if booking.duration > mm:
                    raise Exception("The duration of the booking is greater that "
                                    "the maximum allowed for the resource. ")

            # Validate if there are restrictions in max number of bookings for
            # this type of resource or similar ones (same tags)
            pending = self.__get_session_dict('pending_bookings')
            nextBookings = None
            for tagName, maxPending in pending.items():
                m = int(maxPending)
                if m > 0 and tagName in r.tags:
                    # Only retrieve the next bookings when it is required
                    if nextBookings is None:
                        nextBookings = self.get_next_bookings(user)
                    count = sum(1 for b in nextBookings if tagName in b.resource.tags)
                    if count >= m:
                        raise Exception("You already reached the maximum number"
                                        " of pending bookings for resource tag "
                                        "'%s'" % tagName)

        overlap = self.get_bookings_range(booking.start,
                                          booking.end,
                                          resource=r)

        app = None

        if not booking.is_slot:
            # Check there is not overlapping with other non-slot events
            overlap_noslots = [b for b in overlap
                               if not b.is_slot and b.id != booking.id]
            if overlap_noslots:
                raise Exception("Booking is overlapping with other events: %s"
                                % overlap_noslots)

            overlap_slots = [b for b in overlap
                             if b.is_slot and b.id != booking.id]

            # Always try to find the Application to set in the booking unless
            # the owner is a manager
            owner = self.get_user_by(id=booking.owner_id)

            if not owner.is_manager:
                apps = owner.get_applications()
                n = len(apps)

                if n == 0 and r.requires_application:
                    raise Exception("User %s has no active application"
                                    % owner.name)

                # Let's try to find an application that allows the owner to book
                def find_app():
                    for b in overlap_slots:
                        for a in apps:
                            if b.application_in_slot(a):
                                return a

                    for a in apps:
                        if a.no_slot(r.id):
                            return a

                    return None

                app = find_app()

                # In the case of a manager updating a booking, the manager
                # can do the booking despise the SLOTs and Application rules
                if app is None and user.is_manager and apps:
                    app = apps[0]

                user_can_book = any(user.can_book_slot(s) for s in overlap_slots)

                if (app is None and not user.is_manager
                    and not user_can_book and r.requires_slot):
                    raise Exception("You do not have permission to book "
                                    "outside slots for this resource or have not "
                                    "access to the given slot. ")

        if app is not None:
            booking.application_id = app.id
            count = self.count_booking_resources([app.id],
                                                 resource_tags=r.tags.split())
            for tagKey, tagCount in count[app.id].items():
                alloc = app.get_quota(tagKey)
                if alloc:  # if different from None or 0, then check
                    if tagCount + booking.days > alloc:
                        raise Exception("Exceeded number of allocated days "
                                        "for application %s on resource tag '%s'"
                                        % (app.code, tagKey))
        else:
            booking.application_id = None

    def __check_cancellation(self, booking, attrs=None):
        """ Check if this booking can be updated or deleted.
        Normal users can only delete or modify the booking up to X hours
        before the starting time. The amount of hours is defined by the
        booking latest_cancellation property.
        Managers can change bookings even the same day and only
        Administrators can change past events.
        This function will raise an exception if a condition is not meet.
        """
        user = self._user
        if user.is_admin:
            return  # admin can cancel/modify at any time

        now = self.now()
        latest = booking.resource.latest_cancellation

        def _error(action):
            raise Exception('This booking can not be %s. \n'
                            'Should be %d hours in advance. '
                            % (action, latest))

        start, end = booking.start, booking.end
        if (not self._user.is_manager
            and start - dt.timedelta(hours=latest) < now):
            if attrs:  # Update case, where we allow modification except dates
                if start != attrs['start'] or end != attrs['end']:
                   _error('updated')
            else:  # Delete case
                _error('deleted')

    def _modify_bookings(self, attrs, modifyFunc):
        """ Return one or many bookings if repeating event.
        Params:
            attrs:
                id: Id of the main booking
                modify_all: If True, all repeating bookings from this one will be
                returned
        """
        booking_id = attrs['id']
        modify_all = attrs.pop('modify_all', False)

        # Get the booking with the given id
        bookings = self.get_bookings(condition='id="%s"' % booking_id)

        if not bookings:
            raise Exception("There is no booking with ID=%s" % booking_id)

        booking = bookings[0]
        rid = booking.repeat_id

        result = [booking]

        if rid is not None:
            repeats = [
                b for b in self.get_bookings(condition='repeat_id="%s"' % rid)
                if b.start > booking.start
            ]
            if modify_all:
                result.extend(repeats)
            else:
                # If not modify_all, we should detach this booking from the
                # repeating series by setting its repeat_id to None
                # and generating a new repeat_id for the future events of the
                # series
                booking.repeat_id = None
                uid = str(uuid.uuid4())
                for b in repeats:
                    b.repeat_id = uid

        for b in result:
            modifyFunc(b)

        self.commit()

        return result

    def _session_data_path(self, session):
        return os.path.join(self._sessionsPath, session.data_path)


class RepeatRanges:
    """ Helper class to generate a series of events with start, end. """
    OPTIONS = {'weekly': 7, 'bi-weekly': 14}

    def __init__(self, frequency, attrs):
        days = self.OPTIONS.get(frequency, None)

        if days is None:
            raise Exception("Invalid repeat value '%s'" % frequency)

        self._delta = dt.timedelta(days=days)
        self._attrs = attrs

    def move(self):
        """ Increase the start, end range by the interal delta. """
        self._attrs['start'] += self._delta
        self._attrs['end'] += self._delta




# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import os

import flask
from flask import request
from flask import current_app as app


pages_bp = flask.Blueprint('pages', __name__)


@pages_bp.route("/", methods=['GET', 'POST'])
def index():
    if flask.request.method == 'GET':
        kwargs = flask.request.args.to_dict()
    else:
        kwargs = flask.request.form.to_dict()

    page_id = kwargs['page_id']

    params = {
        'page_id': page_id,
    }

    # Following params are required in main.html but are not really used
    # here in pages, maybe we should consider to have a separate template
    kwargs = {
        'possible_owners': [],
        'possible_operators': [],
        'resources': [],
    }

    if app.user.is_authenticated:
        app.user.image = app.dc.user_profile_image(app.user)

    return flask.render_template('main.html',
                                 content_id='pages',
                                 params=params,
                                 is_devel=app.is_devel,
                                 version=app.version,
                                 emhub_title=app.config.get('EMHUB_TITLE', ''),
                                 **kwargs)





# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import os
import datetime as dt
import decimal

import sqlalchemy
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.ext.declarative import declarative_base


class DbManager:
    """ Helper class to deal with DB stuff
    """
    def init_db(self, dbPath, cleanDb=False, create=True):
        do_echo = os.environ.get('SQLALCHEMY_ECHO', '0') == '1'

        if cleanDb and os.path.exists(dbPath):
            os.remove(dbPath)

        engine = sqlalchemy.create_engine('sqlite:///' + dbPath, echo=do_echo)

        self._db_session = scoped_session(sessionmaker(autocommit=False,
                                                       autoflush=False,
                                                       bind=engine))
        self.Base = declarative_base()
        self.Base.query = self._db_session.query_property()

        self._create_models()

        # Create the database if it does not exists
        if not os.path.exists(dbPath) and create:
            self.Base.metadata.create_all(bind=engine)

    def commit(self):
        self._db_session.commit()

    def delete(self, item, commit=True):
        self._db_session.delete(item)
        if commit:
            self.commit()

    def close(self):
        self._db_session.remove()

    # ------------------- Some utility methods --------------------------------
    def now(self):
        from tzlocal import get_localzone
        # get local timezone
        local_tz = get_localzone()
        return dt.datetime.now(local_tz)

    @staticmethod
    def json_from_value(v):
        if isinstance(v, dt.date) or isinstance(v, dt.datetime):
            return v.isoformat()
        elif isinstance(v, decimal.Decimal):
            return float(v)
        else:
            return v

    @staticmethod
    def json_from_object(obj):
        """ Return row info as json dict. """
        return {c.key: DbManager.json_from_value(getattr(obj, c.key))
                for c in obj.__table__.c}

    @staticmethod
    def json_from_dict(d):
        """ Return row info as json dict. """
        return {k: DbManager.json_from_value(v) for k, v in d.items()}



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

""" 
This script will check for actions to be taken on sessions,
e.g: create folders or the README file
"""

import os
import sys
import time
import argparse
import subprocess
import datetime as dt
import tempfile
from pprint import pprint


from emhub.client import open_client, config


def create_session_folder(session):
    """ Create the session folder, the session counter might change if the given
    one already exist in the filesystem. In that case the next counter will be
    found. A README file will also be created with some session info.
    """
    session_info = {
        'id': session['id'],
        'status': 'created',
    }

    folder = session['folder']
    name = session['name']
    # We assume that the session counter are the last characters of the name
    # i.g fac00034, dbb00122 or cem00378_00012
    if '_' in name:
        prefix, counterStr = name.split('_')
        counter = int(counterStr)
        sep = '_'
    else:
        prefix = name[:3]
        counter = int(name[3:])
        sep = ''

    def _folderName():
        return '%s%s%05d' % (prefix, sep, counter)

    def _folderPath():
        return os.path.join(folder, _folderName())

    def _run(args):
        print("Running: ", args)
        process = subprocess.run(args, capture_output=True, text=True)
        if process.returncode != 0:
            raise Exception(process.stderr)
        return process

    while os.path.exists(_folderPath()):
        counter += 1

    folderPath = _folderPath()

    try:
        # Allow to define the command to create sessions
        sudo = os.environ.get('EMHUB_SESSION_SUDO', '').split()

        # Create the folder
        _run(sudo + ['mkdir', folderPath])

        session_info['name'] = _folderName()
        session_info['extra'] = {'data_folder': folderPath}

        dateStr = dt.datetime.now().strftime('%Y%m%d')
        readmeFn = os.path.join(folderPath, 'README_%s.TXT' % dateStr)

        f = tempfile.NamedTemporaryFile('w', delete=False)
        for user in ['pi', 'user', 'operator']:
            u = session[user] or {'name': '', 'email': ''}
            f.write('%s.name: %s\n' % (user, u['name']))
            f.write('%s.email: %s\n' % (user, u['email']))
        f.write('description: %s\n' % session['title'])
        f.write('date: %s\n' % dateStr)
        f.close()

        # Move the README file to the folder
        os.chmod(f.name, 0o644)
        _run(sudo + ['cp', f.name, readmeFn])
        os.remove(f.name)

        adduserCmd = os.environ.get('EMHUB_SESSION_ADDUSER', '')

        if adduserCmd:
            args = adduserCmd.split() + [session_info['name']]
            # Add new user to data download machine
            process = _run(args)
            for line in process.stdout.split('\n'):
                if 'Error: ' in line:
                    raise Exception(line)
                elif 'user.password' in line:
                    password = line.split()[1].strip()
                    session_info['extra']['data_user_password'] = password

    except Exception as e:
        session_info['status'] = 'failed'
        session_info['extra'] = {'status_info': str(e)}

    return session_info


def eprint(*args, **kwargs):
    """ print to stderr """
    print(*args, file=sys.stderr, **kwargs)

def main():
    parser = argparse.ArgumentParser()
    add = parser.add_argument  # shortcut

    add('--list', action='store_true',
        help="List existing sessions in the server.")

    args = parser.parse_args()

    # Testing function to just list the sessions
    if args.list:
        with open_client() as dc:
            r = dc.request('get_sessions', jsonData={})
            eprint("Sessions: ")
            for s in r.json():
                eprint("   ", s['name'])
        return

    while True:
        try:
            with open_client() as dc:
                eprint("Connected to server: ", config.EMHUB_SERVER_URL)
                r = dc.request('poll_sessions', jsonData={})

                for s in r.json():
                    eprint("Handling session %s: " % s['id'])
                    eprint("   - Creating folder: ",
                          os.path.join(s['folder'], s['name']))
                    eprint("   - Updating session")
                    session_info = create_session_folder(s)
                    pprint(session_info)
                    dc.update_session(session_info)
        except Exception as e:
            eprint("Some error happened: ", str(e))
            eprint("Waiting 60 seconds before retrying...")
            time.sleep(60)


if __name__ == '__main__':
    main()



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import os
import io
import base64
from PIL import Image, ImageEnhance, ImageOps, ImageFilter

import flask
from flask import request
from flask import current_app as app

from emhub.utils import send_json_data


images_bp = flask.Blueprint('images', __name__)


@images_bp.route("/static", methods=['GET', 'POST'])
def static():
    try:
        fn = request.args['filename']
        return app.send_static_file(os.path.join('images', fn))
    except FileNotFoundError:
        flask.abort(404)


@images_bp.route("/user_profile", methods=['GET', 'POST'])
def user_profile():
    try:
        user_id = request.args['user_id']
        user = app.dm.get_user_by(id=user_id)

        if user.profile_image is None:
            return app.send_static_file(os.path.join('images', 'user-icon.png'))

        return flask.send_from_directory(app.config["USER_IMAGES"],
                                         filename=user.profile_image)
    except FileNotFoundError:
        flask.abort(404)


@images_bp.route("/get_mic_data", methods=['POST'])
def get_mic_data():
    micId = int(request.form['micId'])
    sessionId = int(request.form['sessionId'])
    session = app.dm.load_session(sessionId)
    micSetId = None
    for s in session.data.get_sets():
        if s['id'].startswith('Micrographs'):
            micSetId = s['id']

    if micSetId is None:
        raise Exception("Not micrograph set found in '%s'"
                        % session.data_path)
    attrs = [
        'micThumbData', 'psdData', 'shiftPlotData',
        'ctfDefocusU', 'ctfDefocusV', 'ctfResolution',
        'coordinates', 'micThumbPixelSize', 'pixelSize'
    ]

    mic = session.data.get_set_item(micSetId, micId, attrList=attrs)

    if 'coordinates' in mic:
        mic['coordinates'] = mic['coordinates'].tolist()
    else:
        mic['coordinates'] = []

    def _enhance(base64Str, cutoff=2, radius=1):
        msg = base64.b64decode(base64Str)
        # msg = mic['micThumbData']
        buf = io.BytesIO(msg)
        img = Image.open(buf)
        img = ImageOps.autocontrast(img, cutoff=2)
        img = img.filter(ImageFilter.GaussianBlur(radius=1))
        # enhancer = ImageEnhance.Brightness(img)
        # img = enhancer.enhance(0.15)
        img_io = io.BytesIO()
        img.save(img_io, format='PNG')
        return base64.b64encode(img_io.getvalue()).decode("utf-8")

    config = {
        'micThumbData': {'cutoff': 2, 'radius': 1},
        'psdData': {'cutoff': 0.5, 'radius': 1}
    }
    for key, args in config.items():
        if key in mic:
            mic[key] = _enhance(mic[key], **args)

    session.data.close()

    return send_json_data(mic)



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import json
import datetime as dt
import numpy as np

from . import image


def pretty_json(d):
    print(json.dumps(d, indent=4))


def pretty_date(input_dt):
    if input_dt is None:
        return 'None'

    if isinstance(input_dt, str):
        input_dt = datetime_from_isoformat(input_dt)

    return input_dt.strftime("%Y/%m/%d")


def pretty_datetime(input_dt):
    if input_dt is None:
        return 'None'

    if isinstance(input_dt, str):
        input_dt = datetime_from_isoformat(input_dt)

    return input_dt.strftime("%Y/%m/%d %I:%M %p")


def pretty_quarter(quarter):
    qs, qe = quarter
    return '%s - %s' % (qs.strftime('%b %Y'),
                        qe.strftime('%b %Y'))


def datetime_from_isoformat(iso_string):
    """ Parse the input string and handle ending Z and assume UTC. """
    dt_string = iso_string.replace('Z', '+00:00').replace('+0000', '+00:00')

    return dt.datetime.fromisoformat(dt_string).replace(tzinfo=dt.timezone.utc)


def datetime_to_isoformat(input_dt):
    return input_dt.isoformat().replace('+00:00', 'Z')


def get_quarter(date=None):
    """ Return the quarter date range of a given date
    (now if not date is passed. )
    """
    # If date range is not passed, let's use by default the
    # current quarter
    d = date or dt.datetime.now()
    y = d.year
    qi = (d.month - 1) // 3
    start, end = [
        ('01/01', '03/31'),
        ('04/01', '06/30'),
        ('07/01', '09/30'),
        ('10/01', '12/31')
    ][qi]

    def _dt(value):
        value_str = '%d/%s' % (y, value)
        return dt.datetime.strptime(value_str, '%Y/%m/%d')

    return _dt(start), _dt(end)


class NpJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpJsonEncoder, self).default(obj)


def send_json_data(data):
    import flask
    resp = flask.make_response(json.dumps(data, cls=NpJsonEncoder))
    resp.status_code = 200
    resp.headers['Access-Control-Allow-Origin'] = '*'
    return resp


def send_error(msg):
    return send_json_data({'error': msg})



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

from sqlalchemy import Column, Integer, String, JSON
from sqlalchemy_utc import UtcDateTime

from .data_db import DbManager


class DataLog(DbManager):
    """ Main class that will manage the logs about data operations.
    """
    def __init__(self, dbPath, cleanDb=False):
        self.init_db(dbPath, cleanDb=cleanDb)

    def _create_models(self):
        """ Function called from the init_db method. """

        class Log(self.Base):
            """Model for user accounts."""
            __tablename__ = 'logs'

            id = Column(Integer,
                        primary_key=True)

            user_id = Column(Integer,
                             nullable=True)

            # type log
            type = Column(String(16),
                          nullable=False)

            name = Column(String(256),
                          nullable=False)

            timestamp = Column(UtcDateTime,
                               nullable=False)

            # positional args
            args = Column(JSON, default=[])

            # keywork args
            kwargs = Column(JSON, default={})

        self.Log = Log

    def log(self, log_user_id, log_type, log_name,
            *args, **kwargs):

        log = self.Log(
            user_id=log_user_id,
            type=log_type,
            name=log_name,
            timestamp=self.now(),
            args=args,
            kwargs=kwargs)

        self._db_session.add(log)
        self.commit()

        return log

    def get_logs(self):
        return self._db_session.query(self.Log).all()




# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import os
from collections import namedtuple
import numpy as np
import h5py
import sqlite3
import tables as tbl

from emhub.utils import image


class SessionData:
    """
    Class that will handle the underlying data associate with a given Session.
    It will store information of the acquisition as well as the pre-processing.
    """

    def get_sets(self, attrList=None, condition=None):
        """ Get a list with all or some sets in the session.

        Args:
            attrList: An optional list of attributes, to avoid returning
                all properties for each set. If None, only id's will be returned.
            condition: An optional condition string to filter out the result.
        """
        pass

    def create_set(self, setId, attrsDict):
        """ Create a new set in the session.

        Args:
            setId: The id of the new set that will be created.
            attrsDict: Dict-like object with keys and values

        Return:
            True if the set was successfully created, False otherwise.
        """
        pass

    def update_set(self, setId, attrsDict):
        """ Update existing set attributes.

        Args:
            setId: The id of the new set that will be created.
            attrsDict: Dict-like object with keys and values

        Return:
            True if the set was successfully created, False otherwise.
        """
        pass

    def get_set_items(self, setId, attrList=None, condition=None):
        """ Return a list with all or some items from this set.

        Args:
            setId: The id of the set containing the items.
            attrList: An optional list of attributes, to avoid returning
                all properties for each set. (e.g 'id')
            condition: An optional condition string to filter out
                the result list of objects
        Return:
            A list with items (dict objects)
        """
        pass

    def get_set_item(self, setId, itemId, attrList=None):
        pass

    def add_set_item(self, setId, itemId, attrDict):
        pass

    def update_set_item(self, setId, itemId, attrDict):
        pass


class H5SessionData(SessionData):
    """
    Container of Session Data based on HDF5 file.
    """
    def __init__(self, h5File, mode='r'):
        #h5py.get_config().track_order = True
        print("H5SessionData: mode: ", mode)

        if mode == 'r':
            print("Reading file: ", h5File)
        elif mode in ['w', 'a']:
            os.makedirs(os.path.dirname(h5File), exist_ok=True)
            print("Writing file: ", h5File)

        self._file = h5py.File(h5File, mode)

    def get_sets(self, attrList=None, condition=None):
        setList = []
        for k, v in self._file[self._getSetPath('')].items():
            if not attrList:
                setAttrs = dict(v.attrs)
            else:
                setAttrs = {a: v.attrs[a] for a in attrList}
            setAttrs['id'] = k
            setList.append(setAttrs)

        return setList

    def _set_group_attrs(self, group, setId, attrDict):
        attrs = {'id': setId}
        attrs.update(attrDict)
        for k, v in attrs.items():
            group.attrs[k] = v

    def create_set(self, setId, attrDict):
        group = self._file.create_group(self._getSetPath(setId))
        self._set_group_attrs(group, setId, attrDict)

    def update_set(self, setId, attrDict):
        group = self._file[self._getSetPath(setId)]
        self._set_group_attrs(group, setId, attrDict)

    def get_set_item(self, setId, itemId, attrList=None):
        itemAttrs = self._file[self._getItemPath(setId, itemId)].attrs
        return {a: itemAttrs[a] for a in attrList if a in itemAttrs}

    def get_set_items(self, setId, attrList=None, condition=None):
        # print(">>> Getting items from ", self._getSetPath(setId))
        # print(self.get_sets())

        if attrList is None:
            attrs = list(ImageSessionData.MIC_ATTRS.keys())
        elif 'id' not in attrList:
            attrs = ['id'] + attrList
        else:
            attrs = attrList

        # Check that all requested attributes in attrList are valid for Micrograph
        # if any(a not in ImageSessionData.MIC_ALL_ATTRS for a in attrs):
        #     raise Exception("Invalid attribute for micrograph")

        itemsList = []

        setGroup = self._file[self._getSetPath(setId)]

        for item in setGroup.values():
            #print("  item: ", item.name)
            values = {a: item.attrs[a] for a in attrs if a in item.attrs}
            itemsList.append(values)

        return itemsList

    def add_set_item(self, setId, itemId, attrDict):
        micGroup = self._file.create_group(self._getItemPath(setId, itemId))
        micAttrs = micGroup.attrs
        micAttrs['id'] = itemId

        for key, value in attrDict.items():
            # try:
                if isinstance(value, np.ndarray):
                    micGroup.create_dataset(key, data=value)
                else:
                    micAttrs[key] = value
            # except:
            #     if value is not None:
            #         #micGroup.create_dataset(key, data=value)
            #         print("Setting attribute: ", key)
            #         print("   type: ", type(value))
            #         print("   >>> Value is None")

    def update_set_item(self, setId, itemId, attrDict):
        print("update_set_item:  Getting H5 path: " + self._getItemPath(setId, itemId));
        micAttrs = self._file[self._getItemPath(setId, itemId)].attrs
        micAttrs.update(**attrDict)

    def close(self):
        self._file.close()

    def _getItemPath(self, setId, itemId):
        return '%s/item%06d' % (self._getSetPath(setId), itemId)

    def _getSetPath(self, setId):
        return '/Sets/%s' % setId


class ImageSessionData(SessionData):
    """
    Very simple implementation of SessionData for testing purposes.
    This class depends on the definition of EMHUB_TESTDATA environment
    variable and the t20s_pngs folder inside it.
    """

    MIC_ATTRS = {
        'id': 'id',
        'location': 'c11',
        'ctfDefocus': 'c01',
        'ctfDefocusU': 'c01',
        'ctfDefocusV': 'c02',
        'ctfDefocusAngle': 'c03',
        'ctfResolution': 'c06',
        'ctfFit': 'c07'
    }

    MIC_DATA_ATTRS = [
        'micThumbData', 'psdData', 'ctfFitData', 'shiftPlotData'
    ]

    MIC_EXTRA_ATTRS = [
        'coordinates'
    ]

    MIC_ALL_ATTRS = list(MIC_ATTRS.keys()) + MIC_DATA_ATTRS + MIC_EXTRA_ATTRS

    def __init__(self, useBase64=True):
        self._useBase64 = useBase64

        testData = os.environ.get('EMHUB_TESTDATA', None)

        if testData is None:
            raise Exception('EMHUB_TESTDATA not defined.')

        self.dataDir = os.path.abspath(os.path.join(testData, 't20s_pngs'))

        if not os.path.exists(self.dataDir):
            raise Exception("Missing %s, defined by EMHUB_TESTDATA"
                            % self.dataDir)

        self._rows = self.get_micrograph_rows(self.getFile('ctfs.sqlite'))
        self._rowsDict = {row['id']: row for row in self._rows}

    def getFile(self, *paths):
        return os.path.join(self.dataDir, *paths)

    def get_sets(self, attrList=None, condition=None):
        return [{'id': 1}]

    def create_set(self, setId, attrDict):
        raise Exception("Not supported.")

    def get_set_items(self, setId, attrList=None, condition=None):
        if condition is not None:
            raise Exception("condition evaluation not implemented. ")

        attrs = self._get_attrs(attrList)

        return [self._get_dict_from_row(row, attrs) for row in self._rows]

    def get_set_item(self, setId, itemId, attrList=None):
        return self._get_dict_from_row(self._rowsDict[itemId],
                                       self._get_attrs(attrList))

    def add_set_item(self, setId, attrDict):
        raise Exception("Not supported.")

    def update_set_item(self, setId, attrDict):
        raise Exception("Not supported.")

    # ------------------------ Utility functions ------------------------
    def _load_image(self, fn):
        if self._useBase64:
            return image.fn_to_base64(fn)
        else:
            return image.fn_to_blob(fn)

    def compute_micThumbData(self, micPrefix):
        fn = self.getFile('imgMic/%s_thumbnail.png' % micPrefix)
        return self._load_image(fn)

    def compute_psdData(self, micPrefix):
        fn = self.getFile('imgPsd/%s_aligned_mic_ctfEstimation.png'
                          % micPrefix)
        return self._load_image(fn)

    def compute_shiftPlotData(self, micPrefix):
        return self._load_image(self.getFile('imgShift/%s_global_shifts.png'
                                             % micPrefix))

    def get_micrograph_rows(self, sqliteFn, micId=None):
        """ Load a sqlite file produced by Scipion. """
        rows = None
        try:
            conn = sqlite3.connect(sqliteFn)
            conn.row_factory = sqlite3.Row
            cur = conn.cursor()
            whereStr = '' if micId is None else ' WHERE id=%s' % micId
            cur.execute("SELECT * FROM Objects" + whereStr)
            rows = cur.fetchall()
            conn.close()
        except Exception as e:
            print(e)

        return rows

    def _get_attrs(self, attrList):
        if attrList is None:
            attrs = list(self.MIC_ATTRS.keys())
        elif 'id' not in attrList:
            attrs = ['id'] + attrList
        else:
            attrs = attrList

        # Check that all requested attributes in attrList are valid for Micrograph
        if any(a not in self.MIC_ALL_ATTRS for a in attrs):
            raise Exception("Invalid attribute for micrograph")

        return attrs

    def _get_dict_from_row(self, row, attrs):
        values = {a: row[self.MIC_ATTRS[a]] for a in attrs if a in self.MIC_ATTRS}

        micPrefix = os.path.basename(values['location']).replace('_aligned_mic.mrc', '')
        for a in attrs:
            # Compute data for some of the attributes
            computeFunc = getattr(self, 'compute_%s' % a, None)
            if computeFunc:
                values[a] = computeFunc(micPrefix)

        return values


class PytablesSessionData(SessionData):
    """ Implementation of SessionData using pytables. """
    #TODO: pytables stores strings as bytes, need to decode when reading
    # they are also fixed size length :(

    class SetOfMicrographs(tbl.IsDescription):
        id = tbl.Int32Col()
        extra = tbl.StringCol(256)

    class Micrograph(tbl.IsDescription):
        id = tbl.Int32Col()
        location = tbl.StringCol(256)
        ctfDefocus = tbl.Float32Col()
        ctfDefocusU = tbl.Float32Col()
        ctfDefocusV = tbl.Float32Col()
        ctfDefocusAngle = tbl.Float32Col()
        ctfResolution = tbl.Float32Col()
        ctfFit = tbl.Float32Col()
        micThumbData = tbl.StringCol(256000)
        psdData = tbl.StringCol(256000)
        ctfFitData = tbl.StringCol(256000)
        shiftPlotData = tbl.StringCol(256000)

    def __init__(self, h5File, mode='r'):
        if mode == 'r':
            print("Reading file: ", h5File)
        elif mode in ['w', 'a']:
            os.makedirs(os.path.dirname(h5File), exist_ok=True)
            print("Writing file: ", h5File)

        self._file = tbl.open_file(h5File, mode)

    def get_sets(self, attrList=None, condition=None, setId=None):
        setList = []
        if setId is not None:
            setTbl = self._file.get_node(self._getMicSet(setId) + '/set_tbl')
            # setTbl has only one row
            row = setTbl[0]
            setList.append({k: row[k] for k in setTbl.colnames})
        else:
            for grp in self._file.iter_nodes("/Micrographs"):
                values = {k: grp.set_tbl[0][k] for k in grp.set_tbl.colnames}
                setList.append(values)

        return setList

    def create_set(self, setId, attrs):
        self._file.create_group("/Micrographs", "set%03d" % setId,
                                createparents=True)
        setTbl = self._file.create_table(self._getMicSet(setId), 'set_tbl',
                                         self.SetOfMicrographs,
                                         "Set table")
        setRow = setTbl.row
        attrs.update({'id': setId})

        for key, value in attrs.items():
            setRow[key] = value
        setRow.append()
        setTbl.flush()
        mics_table = self._file.create_table(self._getMicSet(setId), 'mics_tbl',
                                             self.Micrograph, "Mics table")
        mics_table.flush()

    def get_items(self, setId, attrList=None, condition=None,
                  itemId=None):
        mics_table = self._file.get_node(self._getMicSet(setId), 'mics_tbl')

        if attrList is None:
            keys = mics_table.colnames
        elif 'id' not in attrList:
            keys = ['id'] + attrList
        else:
            keys = attrList

        Micrograph = namedtuple('Micrograph', keys)
        micList = []

        for row in mics_table:
            values = dict()
            for k in keys:
                if isinstance(row[k], bytes):
                    values[k] = row[k].decode()
                else:
                    values[k] = row[k]

            #values = {k: row[k] for k in keys}
            micList.append(Micrograph(**values))

        return micList

    def get_item(self, setId, itemId, dataAttrs=None):
        print("Requesting item: setId: %s, itemId: %s" % (setId, itemId))
        mics_table = self._file.get_node(self._getMicSet(setId), 'mics_tbl')
        mic = mics_table[itemId-1]  # pytables rows start from 0

        keys = mics_table.colnames if dataAttrs is None else dataAttrs
        Micrograph = namedtuple('Micrograph',keys)
        values = dict()
        for k in keys:
            if isinstance(mic[k], bytes):
                values[k] = mic[k].decode()
            else:
                values[k] = mic[k]

        #values = {k: mic[k] for k in keys}
        return Micrograph(**values)

    def add_item(self, setId, itemId, **attrDict):
        mics_table = self._file.get_node(self._getMicSet(setId), 'mics_tbl')
        mic = mics_table.row

        attrDict.update({'id': itemId})
        for key, value in attrDict.items():
            mic[key] = value

        mic.append()
        mics_table.flush()

    def update_item(self, setId, itemId, **attrDict):
        mics_table = self._file.get_node(self._getMicSet(setId), 'mics_tbl')
        itemId = int(itemId) - 1  # pytables rows start from 0
        mics_table.modify_columns(itemId, itemId+1,
                                  columns=[[x] for x in attrDict.values()],
                                  names=list(attrDict.keys()))
        mics_table.flush()

    def close(self):
        self._file.close()

    def _getMicSet(self, setId):
        return '/Micrographs/set%03d' % setId
